from typing import AnyStr
from typing import Callable
from typing import Optional
from typing import Union
from typing import Tuple
from typing import List
from typing import Dict
from typing import Any
from typing_extensions import Literal

import sys
import os
import json
import tempfile
import shutil

from functools import partial

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets
from qtpy.QtGui import QIcon

from compas.data import Data
from compas.colors import Color
from compas.utilities import gif_from_images

from compas_view2.views import View120
from compas_view2.views import View330
from compas_view2.objects import Object
from compas_view2.forms.dockform import DockForm
from compas_view2.forms.sceneform import SceneForm
from compas_view2.forms.propertyform import PropertyForm
from compas_view2.forms.treeform import TreeForm
from compas_view2.forms.tabsform import TabsForm

from compas_view2.ui import Button
from compas_view2.ui import Slider
from compas_view2.ui import Radio
from compas_view2.ui import Checkbox
from compas_view2.ui import Select

try:
    from compas_view2.flow import Flow
except ImportError:
    Flow = None

from .timer import Timer
from .selector import Selector
from .controller import Controller
from .worker import Worker
from .plot import MplCanvas

HERE = os.path.dirname(__file__)
ICONS = os.path.join(HERE, "../icons")
CONFIG = os.path.join(HERE, "config.json")

VERSIONS = {"120": (2, 1), "330": (3, 3)}


class App:
    """Viewer app.

    Parameters
    ----------
    title : str, optional
        The title of the viewer window.
    version: {'120', '330'}, optional
        The version of the GLSL used by the shaders.
        Default is ``'120'`` with a compatibility profile.
        The option ``'330'`` is not yet available.
    width: int, optional
        The width of the app window at startup.
    height: int, optional
        The height of the app window at startup.
    viewmode: {'shaded', 'ghosted', 'wireframe', 'lighted'}, optional
        The display mode of the OpenGL view.
        In `ghosted` mode, all objects have a default opacity of 0.7.
    show_grid: bool, optional
        Show the XY plane.
    controller_class: :class:`compas_view2.app.Controller`, optional
        A custom controller corresponding to a custom config file.
        Default is None, in which case the default controller is used, matching the default config file.
    config: dict | filepath, optional
        A configuration dict for the UI, or a path to a JSON file containing such a dict.
        Default is None, in which case the default configuration is used.

    Attributes
    ----------
    window: :class:`PySide2.QtWidgets.QMainWindow`
        The main window of the application.
        This window contains the view and any other UI components
        such as the menu, toolbar, statusbar, ...
    view: :class:`compas_view2.View`
        Instance of OpenGL view.
        This view is the central widget of the main window.
    controller: :class:`compas_view2.app.Controller`
        The action controller of the app.

    Notes
    -----
    The app has a (main) window with a central OpenGL widget (i.e. the 'view'),
    and a menubar, toolbar, and statusbar.
    The menubar provides access to all supported 'actions'.
    The toolbar is meant to be a 'quicknav' to a selected set of actions.
    The app supports rotate/pan/zoom, and object selection via picking or box selections.

    Currently the app uses OpenGL 2.2 and GLSL 120 with a 'compatibility' profile.
    Support for OpenGL 3.3 and GLSL 330 with a 'core' profile is under development.

    The app can currently only be used 'as-is'.
    This means that there is no formal mechanism for adding actions to the controller
    or to add functionality to the shader, other than by extending the core classes.
    In the future, such mechanism will be provided by allowing the user to overwrite
    the configuration file and add actions to the controller, without having
    to modify the package source code.

    Currently the app has no scene graph.
    All added COMPAS objects are wrapped in a viewer object and stored in a dictionary,
    mapping the object's ID (``id(object)``) to the instance.

    Examples
    --------
    >>> from compas_view2 import app
    >>> viewer = app.App()
    >>> viewer.show()

    """

    def __init__(
        self,
        title: str = "COMPAS View2",
        version: Literal["120", "330"] = "120",
        width: int = 800,
        height: int = 500,
        viewmode: Literal["wireframe", "shaded", "ghosted", "lighted"] = "shaded",
        controller_class: Optional[Controller] = None,
        show_grid: bool = True,
        config: Optional[dict] = None,
        enable_sidebar: bool = False,
        enable_sidedock1: bool = False,
        enable_sidedock2: bool = False,
        enable_sceneform: bool = False,
        enable_propertyform: bool = False,
        show_flow: bool = False,
        flow_view_size: Union[Tuple[int], List[int]] = None,
        flow_auto_update: bool = True,
    ):

        if version not in VERSIONS:
            raise Exception(
                "Only these versions are currently supported: {}".format(VERSIONS)
            )

        glFormat = QtGui.QSurfaceFormat()
        glFormat.setVersion(*VERSIONS[version])

        if version == "330":
            View = View330
            glFormat.setProfile(QtGui.QSurfaceFormat.CoreProfile)
        elif version == "120":
            View = View120
            glFormat.setProfile(QtGui.QSurfaceFormat.CompatibilityProfile)
        else:
            raise NotImplementedError

        glFormat.setDefaultFormat(glFormat)
        QtGui.QSurfaceFormat.setDefaultFormat(glFormat)

        app = QtCore.QCoreApplication.instance()
        if app is None:
            app = QtWidgets.QApplication(sys.argv)
        app.references = set()

        appIcon = QIcon(os.path.join(ICONS, "compas_icon_white.png"))
        app.setWindowIcon(appIcon)
        app.setApplicationName(title)

        self.timer = None
        self.frame_count = 0
        self.record = False
        self.recorded_frames = []

        self.width = width
        self.height = height
        self.window = QtWidgets.QMainWindow()
        self.view = View(self, mode=viewmode, show_grid=show_grid)
        self.window.setCentralWidget(self.view)
        self.window.setContentsMargins(0, 0, 0, 0)

        controller_class = controller_class or Controller
        self.controller = controller_class(self)

        config = config or CONFIG
        if not isinstance(config, dict):
            with open(config) as f:
                config = json.load(f)

        self.config = config

        self._app = app
        self._app.references.add(self.window)
        self.selector = Selector(self)

        self.show_flow = show_flow
        if Flow:
            self.flow = Flow(
                self,
                flow_view_size=flow_view_size or (self.width, self.height),
                flow_auto_update=flow_auto_update,
            )

        self.enable_sidebar = enable_sidebar
        self.enable_sidedock1 = enable_sidedock1
        self.enable_sidedock2 = enable_sidedock2
        self.enable_sceneform = enable_sceneform
        self.enable_propertyform = enable_propertyform
        self.dock_slots = {
            "sceneform": None,
            "propertyform": None,
        }

        self.init()
        self.resize(width, height)
        self.started = False
        self.on_object_selected = []

    def init(self):
        """Initialize the components of the user interface.

        Returns
        -------
        None

        """
        self._init_statusbar()
        self._init_menubar(self.config.get("menubar"))
        self._init_toolbar(self.config.get("toolbar"))
        self._init_sidebar(self.config.get("sidebar"))
        self._init_sidedocks()

    def resize(self, width: int, height: int):
        """Resize the main window programmatically.

        Parameters
        ----------
        width: int
            Width of the viewer window.
        height: int
            Height of the viewer window.

        Returns
        -------
        None

        """
        self.window.resize(width, height)
        desktop = self._app.desktop()
        rect = desktop.availableGeometry()
        x = 0.5 * (rect.width() - width)
        y = 0.5 * (rect.height() - height)
        self.window.setGeometry(x, y, width, height)

    def add(
        self,
        data: Data,
        name: str = None,
        is_selected: bool = None,
        is_visible: bool = None,
        show_points: bool = None,
        show_lines: bool = None,
        show_faces: bool = None,
        pointcolor: Union[Color, Dict[Union[str, int], Color]] = None,
        linecolor: Union[Color, Dict[Union[str, int], Color]] = None,
        facecolor: Union[Color, Dict[Union[str, int], Color]] = None,
        linewidth: int = None,
        pointsize: int = None,
        opacity: int = None,
        **kwargs,
    ) -> Object:
        """Add a COMPAS object.

        Parameters
        ----------
        data: :class:`compas.data.Data`
            A COMPAS data object.
        is_selected : bool, optional
            Whether the object is selected.
            Default to False.
        is_visible : bool, optional
            Whether to show object.
            Default to True.
        show_points : bool, optional
            Whether to show points/vertices of the object.
            Default to False.
        show_lines : bool, optional
            Whether to show lines/edges of the object.
            Default to True.
        show_faces : bool, optional
            Whether to show faces of the object.
            Default to True.
        pointcolor : Union[Color, Dict[Union[str, int], Color]], optional
            The color or the dict of colors of the points.
            Default to `compas_view2.objects.Object.default_color_points`.
        linecolor : Union[Color, Dict[Union[str, int], Color]], optional
            The color or the dict of colors of the lines.
            Default to `compas_view2.objects.Object.default_color_lines`.
        facecolor : Union[Color, Dict[Union[str, int], Color]], optional
            The color or the dict of colors of the faces.
            Default to `compas_view2.objects.Object.default_color_faces`.
        linewidth : int, optional
            The line width to be drawn on screen
            Default to 1.
        pointsize : int, optional
            The point size to be drawn on screen
            Default to 10.
        opacity : float, optional
            The opacity of the object.
            Default to 1.0.
        **kwargs : dict, optional
            Additional visualization options for specific objects.

        Returns
        -------
        :class:`compas_view2.objects.Object`
            The added object.

        """
        if name is not None:
            kwargs["name"] = name
        if is_selected is not None:
            kwargs["is_selected"] = is_selected
        if is_visible is not None:
            kwargs["is_visible"] = is_visible
        if show_points is not None:
            kwargs["show_points"] = show_points
        if show_lines is not None:
            kwargs["show_lines"] = show_lines
        if show_faces is not None:
            kwargs["show_faces"] = show_faces
        if pointcolor is not None:
            kwargs["pointcolor"] = pointcolor
        if linecolor is not None:
            kwargs["linecolor"] = linecolor
        if facecolor is not None:
            kwargs["facecolor"] = facecolor
        if linewidth is not None:
            kwargs["linewidth"] = linewidth
        if pointsize is not None:
            kwargs["pointsize"] = pointsize
        if opacity is not None:
            kwargs["opacity"] = opacity

        obj = Object.build(data, app=self, **kwargs)

        self.view.objects[obj] = obj
        self.selector.add(obj)
        if self.view.isValid():
            obj.init()
            if self.dock_slots["sceneform"]:
                self.dock_slots["sceneform"].update()
        return obj

    def add_reference(self, obj: Object, **kwargs) -> Object:
        """Add an object as a reference to another object.

        Parameters
        ----------
        obj : :class:`compas_view2.objects.Object`
            A view object.
        **kwargs : dict, optional
            Additional visualization options.

        Returns
        -------
        :class:`compas_view2.objects.Object`

        """
        ref = obj.otype.from_other(obj, **kwargs)
        self.view.objects[ref] = ref
        self.selector.add(ref)
        if self.view.isValid():
            ref.init()
        return ref

    def remove(self, obj: Object) -> None:
        """Remove an object from the view.

        Parameters
        ----------
        obj : :class:`compas_view2.objects.Object`
            A view object.

        Returns
        -------
        None

        """
        if obj in list(self.view.objects):
            del self.view.objects[obj]
        for key, value in list(self.selector.instances.items()):
            if obj == value:
                del self.selector.instances[key]

    def show(self) -> None:
        """Show the viewer window.

        Returns
        -------
        None

        """
        self.started = True
        self.window.show()
        if Flow and self.show_flow:
            self.flow.show()

        if self.dock_slots["sceneform"]:
            self.dock_slots["sceneform"].update()

        self._app.exec_()

    run = show

    def about(self) -> None:
        """Display the about message as defined in the config file.

        Returns
        -------
        None

        """
        QtWidgets.QMessageBox.about(
            self.window, "About", self.config["messages"]["about"]
        )

    def info(self, message: str) -> None:
        """Display info.

        Parameters
        ----------
        message : str
            An info message.

        Returns
        -------
        None

        """
        result = QtWidgets.QMessageBox.information(self.window, "Info", message)
        print(result)

    def warning(self, message: str) -> None:
        """Display a warning.

        Parameters
        ----------
        message : str
            A warning message.

        Returns
        -------
        None

        """
        result = QtWidgets.QMessageBox.warning(self.window, "Warning", message)
        print(result)

    def critical(self, message: str) -> None:
        """Display a critical warning.

        Parameters
        ----------
        message : str
            A critical warning message.

        Returns
        -------
        None

        """
        result = QtWidgets.QMessageBox.critical(self.window, "Critical", message)
        print(result)

    def question(self, message: str) -> None:
        """Ask a question.

        Parameters
        ----------
        message : str
            A question.

        Returns
        -------
        None

        """
        flags = QtWidgets.QMessageBox.StandardButton.Yes
        flags |= QtWidgets.QMessageBox.StandardButton.No
        response = QtWidgets.QMessageBox.question(
            self.window, "Question", message, flags
        )
        if response == QtWidgets.QMessageBox.Yes:
            return True
        return False

    def confirm(self, message: str):
        """Confirm the execution of an action.

        Parameters
        ----------
        message : str
            Message to inform the user.

        Returns
        -------
        bool
            True if the user confirms.
            False otherwise.

        Examples
        --------
        .. code-block:: python

            if viewer.confirm("Should i continue?"):
                continue

        """
        flags = QtWidgets.QMessageBox.StandardButton.Ok
        flags |= QtWidgets.QMessageBox.StandardButton.Cancel
        response = QtWidgets.QMessageBox.warning(
            self.window, "Confirmation", message, flags
        )
        if response == QtWidgets.QMessageBox.StandardButton.Ok:
            return True
        return False

    def status(self, message: str) -> None:
        """Display a message in the status bar.

        Parameters
        ----------
        message : str
            A status message.

        Returns
        -------
        None

        """
        self.statusText.setText(message)

    def fps(self, fps: int) -> None:
        """Update fps info in the status bar.

        Parameters
        ----------
        fps : int
            The number of frames per second.

        Returns
        -------
        None

        """
        self.statusFps.setText("fps: {}".format(fps))

    def sidedock(self, title: str = "", slot: str = None, location: str = "right"):
        """Create a side dock widget."""
        if slot and slot in self.dock_slots:
            self.dock_slots[slot].close()

        locations = {
            "left": QtCore.Qt.LeftDockWidgetArea,
            "right": QtCore.Qt.RightDockWidgetArea,
            "top": QtCore.Qt.TopDockWidgetArea,
            "bottom": QtCore.Qt.BottomDockWidgetArea,
        }

        dock = DockForm(self, title)
        self.window.addDockWidget(locations[location], dock)

        if slot:
            self.dock_slots[slot] = dock

        return dock

    def sceneform(self):
        """Create a side object tree form widget."""
        if self.dock_slots["sceneform"]:
            self.dock_slots["sceneform"].show()
            return self.dock_slots["sceneform"]

        sceneform = SceneForm(self)
        self.window.addDockWidget(QtCore.Qt.RightDockWidgetArea, sceneform)
        self.dock_slots["sceneform"] = sceneform

        return sceneform

    def propertyform(self):
        """Create a side object tree form widget."""
        if self.dock_slots["propertyform"]:
            self.dock_slots["propertyform"].show()
            return self.dock_slots["propertyform"]

        propertyform = PropertyForm("Properties", on_update=self.view.update)
        self.window.addDockWidget(QtCore.Qt.RightDockWidgetArea, propertyform)
        self.dock_slots["propertyform"] = propertyform

        return propertyform

    def treeform(
        self,
        title="tree",
        data=None,
        slot: str = None,
        location: str = "left",
        floating=False,
        columns=["key", "value"],
        show_headers=True,
        striped_rows=False,
    ):
        """Create a side object tree form widget."""
        if slot and slot in self.dock_slots:
            treeform = self.dock_slots[slot]
            treeform.setWindowTitle(title)
            treeform.update(data)
            return treeform

        locations = {
            "left": QtCore.Qt.LeftDockWidgetArea,
            "right": QtCore.Qt.RightDockWidgetArea,
            "top": QtCore.Qt.TopDockWidgetArea,
            "bottom": QtCore.Qt.BottomDockWidgetArea,
        }

        treeform = TreeForm(
            self,
            title=title,
            data=data,
            columns=columns,
            show_headers=show_headers,
            striped_rows=striped_rows,
        )
        self.window.addDockWidget(locations[location], treeform)

        if slot:
            self.dock_slots[slot] = treeform
        if floating:
            treeform.setFloating(True)

        return treeform

    def tabsform(
        self,
        title="tree",
        tabs=[],
        slot: str = None,
        location: str = "left",
        floating=False,
        columns=["key", "value"],
        show_headers=True,
        striped_rows=False,
    ):
        """Create a tabs form widget."""
        if slot and slot in self.dock_slots:
            tabsform = self.dock_slots[slot]
            tabsform.setWindowTitle(title)
            tabsform.update(tabs)
            return tabsform

        locations = {
            "left": QtCore.Qt.LeftDockWidgetArea,
            "right": QtCore.Qt.RightDockWidgetArea,
            "top": QtCore.Qt.TopDockWidgetArea,
            "bottom": QtCore.Qt.BottomDockWidgetArea,
        }

        tabsform = TabsForm(
            self,
            title=title,
            tabs=tabs,
            columns=columns,
            show_headers=show_headers,
            striped_rows=striped_rows,
        )
        self.window.addDockWidget(locations[location], tabsform)

        if slot:
            self.dock_slots[slot] = tabsform
        if floating:
            tabsform.setFloating(True)

        return tabsform

    def popup(self, title: str = "", slot: str = None):
        """Create a side dock widget."""
        popup = self.sidedock(title, slot)
        popup.setFloating(True)
        return popup

    def plot(
        self,
        title: str = "",
        location: str = "bottom",
        floating: bool = False,
        min_height: int = 200,
        min_width: int = 200,
    ):
        """Create a matplotlib canvas as dock widget."""
        dock = self.sidedock(title, location=location)
        dock.setFloating(floating)
        dock.setMinimumHeight(min_height)
        dock.setMinimumWidth(min_width)
        sc = MplCanvas()
        dock.content_layout.addWidget(sc)
        return sc.figure

    def threading(
        self,
        func: Callable,
        args: list = [],
        kwargs: dict = {},
        on_progress: Callable = None,
        on_result: Callable = None,
        include_self: bool = False,
    ) -> None:
        """Execute a multi-threaded function.

        Parameters
        ----------
        func : function
            The function to be executed.
        args : list, optional
            The arguments to be passed to the function.
        kwargs : dict, optional
            The keyword arguments to be passed to the function.
        on_progress : function, optional
            A function to be called on progress event.
        on_result : function, optional
            A function to be called on result event.
        include_self : bool, optional
            Include the thread worker instance in the arguments, for sending out progress singals.

        Returns
        -------
        None

        """
        worker = Worker(func, args=args, kwargs=kwargs, include_self=include_self)
        if on_progress:
            worker.signals.progress.connect(on_progress)
        if on_result:
            worker.signals.result.connect(on_result)
        Worker.pool.start(worker)

    # ==============================================================================
    # UI
    # ==============================================================================

    def _get_icon(self, icon: str):
        return QtGui.QIcon(os.path.join(ICONS, icon))

    def _init_statusbar(self):
        self.statusbar = self.window.statusBar()
        self.statusbar.setContentsMargins(0, 0, 0, 0)
        self.statusText = QtWidgets.QLabel("Ready")
        self.statusbar.addWidget(self.statusText, 1)
        self.statusFps = QtWidgets.QLabel("fps: ")
        self.statusbar.addWidget(self.statusFps)

    def _init_menubar(self, items: List[Dict]):
        if not items:
            return
        self.menubar = self.window.menuBar()
        self.menubar.setNativeMenuBar(True)
        self.menubar.setContentsMargins(0, 0, 0, 0)
        self._add_menubar_items(items, self.menubar)

    def _init_toolbar(self, items: List[Dict]):
        if not items:
            return
        self.toolbar = self.window.addToolBar("Tools")
        self.toolbar.setMovable(False)
        self.toolbar.setObjectName("Tools")
        self.toolbar.setIconSize(QtCore.QSize(16, 16))
        self._add_toolbar_items(items, self.toolbar)

    def _init_sidebar(self, items: List[Dict]):
        if not self.enable_sidebar:
            return
        self.sidebar = QtWidgets.QToolBar(self.window)
        self.window.addToolBar(QtCore.Qt.LeftToolBarArea, self.sidebar)
        self.sidebar.setObjectName("Sidebar")
        self.sidebar.setMovable(False)
        self.sidebar.setIconSize(QtCore.QSize(16, 16))
        self.sidebar.setMinimumWidth(240)
        self._add_sidebar_items(items, self.sidebar)

    def _init_sidedocks(self):
        if self.enable_sidedock1:
            self.sidedock1 = self.sidedock(slot="sidedock1")
            self.sidedock1.setFeatures(QtWidgets.QDockWidget.NoDockWidgetFeatures)
        if self.enable_sidedock2:
            self.sidedock2 = self.sidedock(slot="sidedock2")
            self.sidedock2.setFeatures(QtWidgets.QDockWidget.NoDockWidgetFeatures)
        if self.enable_sceneform:
            self.sceneform()
        if self.enable_propertyform:
            self.propertyform()

    def _add_menubar_items(self, items: List[Dict], parent: QtWidgets.QWidget):
        if not items:
            return
        for item in items:
            if item["type"] == "separator":
                parent.addSeparator()
            elif item["type"] == "menu":
                menu = parent.addMenu(item["text"])
                if "items" in item:
                    self._add_menubar_items(item["items"], menu)
            elif item["type"] == "radio":
                radio = QtWidgets.QActionGroup(self.window, exclusive=True)
                for item in item["items"]:
                    action = self._add_action(
                        parent, text=item["text"], action=item["action"]
                    )
                    action.setCheckable(True)
                    action.setChecked(item["checked"])
                    radio.addAction(action)
            elif item["type"] == "action":
                del item["type"]
                self._add_action(parent, text=item["text"], action=item["action"])
            else:
                raise NotImplementedError

    def _add_toolbar_items(self, items: List[Dict], parent: QtWidgets.QWidget):
        if not items:
            return
        for item in items:
            if item["type"] == "separator":
                parent.addSeparator()
            elif item["type"] == "action":
                del item["type"]
                self._add_action(parent, **item)
            else:
                raise NotImplementedError

    def _add_sidebar_items(self, items: List[Dict], parent: QtWidgets.QWidget):
        if not items:
            return
        for item in items:
            if item["type"] == "separator":
                parent.addSeparator()
            elif item["type"] == "radio":
                del item["type"]
                Radio(self, self.sidebar, **item)
            elif item["type"] == "checkbox":
                del item["type"]
                Checkbox(self, self.sidebar, **item)
            elif item["type"] == "slider":
                del item["type"]
                Slider(self, self.sidebar, **item)
            elif item["type"] == "button":
                del item["type"]
                Button(self, self.sidebar, **item)
            else:
                raise NotImplementedError

    def _add_action(
        self,
        parent: QtWidgets.QWidget,
        *,
        text: str,
        action: Callable,
        args: Optional[List[Any]] = None,
        kwargs: Optional[Dict] = None,
        icon: Optional[AnyStr] = None,
    ):
        action = action if callable(action) else getattr(self.controller, action)
        args = args or []
        kwargs = kwargs or {}
        if icon:
            icon = self._get_icon(icon)
            action = parent.addAction(icon, text, partial(action, *args, **kwargs))
        else:
            action = parent.addAction(text, partial(action, *args, **kwargs))
        return action

    # ==============================================================================
    # Decorators
    # ==============================================================================

    def select(self, items: List[Dict[str, Any]], parent=None) -> Callable:
        """Decorator for combo boxes.

        Parameters
        ----------
        items

        Returns
        -------
        callable

        """

        def outer(func: Callable) -> Callable:
            select = Select(self, parent or self.sidebar, items=items, action=func)
            return select

        return outer

    def radio(self, items: List[Dict[str, Any]], title="", parent=None) -> Callable:
        """Decorator for radio actions.

        Parameters
        ----------
        items

        Returns
        -------
        callable

        """

        def outer(func: Callable) -> Callable:
            radio = Radio(
                self, parent or self.sidebar, title=title, items=items, action=func
            )
            return radio

        return outer

    def button(self, text: str, parent=None) -> Callable:
        """Decorator for button actions.

        Parameters
        ----------
        text : str
            The button text label.

        Returns
        -------
        callable

        Notes
        -----
        Use this method to convert a function into the callback action of a button,
        and automatically add the button to the sidebar.

        Examples
        --------
        .. code-block:: python

            @viewer.button('Click me!')
            def click(app):
                app.info('Thanks for clicking...')

        """

        def outer(func: Callable) -> Callable:
            button = Button(self, parent or self.sidebar, text=text, action=func)
            return button

        return outer

    def checkbox(self, text: str, checked: bool = True, parent=None) -> Callable:
        """Decorator for checkbox actions.

        Parameters
        ----------
        text : str
            The text label of the checkbox.
        checked : bool, optional
            If True, the checkbox will be displayed as checked.

        Returns
        -------
        callable

        Notes
        -----
        Use this method to convert a function into the callback action of a checkbox,
        and automatically add the checkbox to the sidebar.

        Examples
        --------
        .. code-block:: python

            @viewer.checkbox('Check me!')
            def check(app):
                app.info('Thanks for checking...')

        """

        def outer(func: Callable) -> Callable:
            checkbox = Checkbox(
                self, parent or self.sidebar, text=text, action=func, checked=checked
            )
            return checkbox

        return outer

    def slider(
        self,
        title: str,
        value: int = 0,
        minval: int = 0,
        maxval: int = 100,
        step: int = 1,
        annotation: str = "",
        bgcolor: Color = None,
        parent=None,
    ) -> Callable:
        """Decorator for slider actions.

        Parameters
        ----------
        title : str
            The text label of the slider.
        value : int, optional
            Initial value of the slider.
        minval : int, optional
            Minimum value of the sliding range.
        maxval : int, optional
            Maximum value of the sliding range.
        step : int, optional
            Size of the sliding step.
        annotation : str, optional
            Value annotation.

        Returns
        -------
        callable

        Notes
        -----
        Use this method to convert a function into the callback action of a slider,
        and automatically add the slider to the sidebar.

        Examples
        --------
        .. code-block:: python

            @viewer.slider(title='Slide me!')
            def slide(app):
                app.info('Thanks for sliding...')

        """

        def outer(func: Callable) -> Callable:
            slider = Slider(
                self,
                parent or self.sidebar,
                func,
                value=value,
                minval=minval,
                maxval=maxval,
                step=step,
                title=title,
                annotation=annotation,
                bgcolor=bgcolor,
            )
            return slider

        return outer

    def on(
        self,
        interval: int = None,
        timeout: int = None,
        frames: int = None,
        record: bool = False,
        record_path: str = "temp/out.gif",
        record_fps: int = None,
        playback_interval: int = None,
    ) -> Callable:
        """Decorator for callbacks of a dynamic drawing process.

        Parameters
        ----------
        interval : int, optional
            Interval between subsequent calls to this function, in milliseconds.
        timeout : int, optional
            Timeout between subsequent calls to this function, in milliseconds.
        frames : int, optional
            The number of frames of the process.
            If no frame number is provided, the process continues until the viewer is closed.
        record : bool, optional
            If True, record a screenshot of every frame.
        record_path : str, optional
            The path where the recording should be saved.
        playback_interval : int, optional
            Interval between frames in the recording, in milliseconds.

        Returns
        -------
        callable

        Notes
        -----
        The difference between `interval` and `timeout` is that the former indicates
        the time between subsequent calls to the callback, without taking into account the duration of the execution of the call,
        whereas the latter indicates a pause after the completed execution of the previous call, before starting the next one.

        Examples
        --------
        .. code-block:: python

            angle = math.radians(5)

            @viewer.on(interval=1000)
            def rotate(frame):
                obj.rotation = [0, 0, frame * angle]
                obj.update()

        """
        if (not interval and not timeout) or (interval and timeout):
            raise ValueError("Must specify either interval or timeout")

        if record:
            # check if record_path is writable
            # create temp dir for frames
            self.tempdir = tempfile.mkdtemp()
            record_fps = record_fps or 1000 / interval

        def outer(func: Callable):
            def render():
                func(self.frame_count)
                self.view.update()
                self.frame_count += 1
                if frames is not None and self.frame_count >= frames:
                    self.timer.stop()
                    if self.record:
                        self.record = False
                        # self.recorded_frames[0].save(
                        #     record_path, save_all=True, optimize=False,
                        #     duration=playback_interval or interval,
                        #     append_images=self.recorded_frames[1:],
                        #     loop=100)
                        files = [
                            os.path.join(self.tempdir, f"{i}.png")
                            for i in range(frames)
                        ]
                        gif_from_images(
                            files=files,
                            gif_path=record_path,
                            fps=record_fps,
                            delete_files=True,
                        )
                        shutil.rmtree(self.tempdir)
                        print("Recorded to ", record_path)

            if interval:
                self.timer = Timer(interval=interval, callback=render)
            if timeout:
                self.timer = Timer(interval=timeout, callback=render, singleshot=True)

            self.frame_count = 0
            self.record = record

        return outer



import ryvencore_qt as rc
import ryvencore
from compas.datastructures import Graph
from typing import Union, Tuple


class Flow(Graph):
    """"A Class that represents maps a Ryven Flow graph.

    Parameters
    ----------
    app : :class:`compas_view2.app.App`
        compas_view app instance.
    flow_auto_update : bool
        Whether to auto update the nodes in graph.
        Defaults to True.
    flow_view_size : Tuple[int, int]
        Window size of the flow view.
        Defaults to (800, 500).

    """

    def __init__(self, app, flow_auto_update: bool = True, flow_view_size: Tuple[int, int] = (800, 500)):
        super().__init__()
        self.app = app
        self.flow_auto_update = flow_auto_update
        self.session = rc.Session()
        self.session.design.set_flow_theme(name='pure dark')
        self.script = self.session.create_script(flow_view_size=flow_view_size)
        self.flow_view = self.session.flow_views[self.script]
        self.script.flow.set_algorithm_mode('data opt')
        self.init_run = False

    def run_all(self):
        """Execute all the ryven nodes in the order of data flow."""
        # print("running all nodes")
        executed = set()
        node_update_states = {node: node.block_updates for node in self.flow_view.node_items}

        def traverse_upwards(node):
            # Traverse upwards to the top of data flow graph
            if node in executed:
                return
            for port in node.inputs:
                for connection in port.connections:
                    traverse_upwards(connection.out.node)
            # print("executing", node)
            node.update_event()
            executed.add(node)

        for node in self.flow_view.node_items:
            node.block_updates = True

        for node in self.flow_view.node_items:
            traverse_upwards(node)

        for node in self.flow_view.node_items:
            node.block_updates = node_update_states[node]
        # print("All nodes executed")

    def show(self):
        """Shows the flow view."""
        if not self.init_run and self.flow_auto_update:
            self.run_all()
            self.init_run = True
        self.flow_view.show()

    def add_node(self, node_class: rc.Node, location: Tuple[int, int] = (0, 0), **kwargs) -> rc.Node:
        """Adds a node to the flow view.

        Parameters
        ----------
        node_class : :class:`ryvencore_qt.Node`
            A ryven node class to be added.
        location : Tuple[int, int]
            The location of the node in the flow view.
            Defaults to (0, 0).
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the node class.

        Returns
        -------
        :class:`ryvencore_qt.Node`
            The created ryven node instance.

        """
        ryven_node = self.script.flow.create_node(node_class, data=kwargs)
        x, y = location
        self.flow_view.node_items[ryven_node].setX(x)
        self.flow_view.node_items[ryven_node].setY(y)
        data = ryven_node.complete_data(ryven_node.data())

        super().add_node(key=ryven_node.GLOBAL_ID, attr_dict={'ryven_data': data})
        return ryven_node

    def add_connection(self, port1: ryvencore.NodePort.NodeOutput, port2: ryvencore.NodePort.NodeInput) -> ryvencore.Connection.DataConnection:
        """Adds a connection between two ryven nodes.

        Parameters
        ----------
        port1 : :class:`ryvencore.NodePort.NodeOutput`
            The first port for the connection.
        port2 : :class:`ryvencore.NodePort.NodeInput`
            The second port for the connection.

        Returns
        -------
        :class:`ryvencore.Connection.DataConnection`
            The created ryven connection instance.

        """
        ryven_connection = self.script.flow.connect_nodes(port1, port2)
        if not ryven_connection:
            return

        # Add connection in compas graph
        node1 = port1.node
        node2 = port2.node
        edge_key = (node1.GLOBAL_ID, node2.GLOBAL_ID)
        if not self.has_edge(*edge_key):
            self.add_edge(*edge_key, {'connections': []})
        connections = self.edge_attribute(edge_key, 'connections')
        connections.append({'port1': self.get_port_info(port1), 'port2': self.get_port_info(port2)})
        self.edge_attribute(edge_key, 'connections', connections)

        return ryven_connection

    def get_port_info(self, port: Union[ryvencore.NodePort.NodeOutput, ryvencore.NodePort.NodeInput]) -> dict:
        """Gets the port information.

        Parameters
        ----------
        port : Union[:class:`ryvencore.NodePort.NodeOutput`, :class:`ryvencore.NodePort.NodeInput`]
            The port to get the information from.

        Returns
        -------
        dict
            The port information.

        """
        node = port.node
        data = {}
        if port in node.inputs:
            data['type'] = 'input'
            data['index'] = node.inputs.index(port)
        elif port in node.outputs:
            data['type'] = 'output'
            data['index'] = node.outputs.index(port)
        else:
            raise ValueError('Port is not in inputs or outputs.')

        data['node'] = node.GLOBAL_ID

        return data



from qtpy import QtWidgets


class Radio:
    """Class representing a radio button group wrapped in a group box.

    Parameters
    ----------
    app : :class:`compas_view2.app.App`
        The app containing the widget.
    parent : QtWidgets.QWidget
        The parent widget for the radio button group.
    title : str
        The label for the radio button group.
    items : list[dict[str, Any]]
        A list of radio button group items, with each item defined as a dict with a particular structure.
        See Notes for more info.
    action : callable
        The action associated with the radio button group.

    Attributes
    ----------
    radio : QtWidgets.QButtonGroup
        The button group containing the exclusive radio items.
    action : callable
        The action associated with the click event of the radio button group.
    values : dict[QtWidgets.QRadioButton, Any]
        Mapping between radio buttons and their values.

    Notes
    -----
    Every `item` dict should have the following structure.

    .. code-block:: python

        item['text']     # str  : The text label of the item.
        item['value']    # Any  : The value of the item, if different from the item text.
        item['checked']  # bool : If True, the item should be marked as checked.

    """

    def __init__(self,
                 app,
                 parent,
                 *,
                 title,
                 items,
                 action):

        box = QtWidgets.QGroupBox(title)
        layout = QtWidgets.QVBoxLayout()
        group = QtWidgets.QButtonGroup(app.window, exclusive=True)
        # add the options
        button_value = {}
        for item in items:
            button = QtWidgets.QRadioButton()
            button.setText(item['text'])
            button.setChecked(item['checked'])
            group.addButton(button)
            layout.addWidget(button)
            button_value[button] = item.get('value', item['text'])
        # connect actions
        group.buttonClicked.connect(self)
        group.buttonClicked.connect(app.view.update)
        # put together
        box.setLayout(layout)
        parent.addWidget(box)
        # define attributes
        self.radio = group
        self.action = action
        self.values = button_value

    def __call__(self, *args, **kwargs):
        """Wrapper for the action associated with the radio button group.

        Returns
        -------
        None

        """
        button = self.radio.checkedButton()
        self.action(self.values[button])



import numpy as np

from compas.utilities import flatten

from compas_view2.gl import make_index_buffer
from compas_view2.gl import make_vertex_buffer
from compas_view2.gl import update_vertex_buffer
from compas_view2.gl import update_index_buffer

from .object import Object


class BufferObject(Object):
    """A shared object to handle GL buffer creation and drawings

    Attributes
    ----------
    visualisation : list[str], read-only
        List of visualisation properties which can be edited in the GUI.
    """

    @property
    def visualisation(self):
        options = ["opacity"]
        if hasattr(self, "_points_data"):
            options += ["pointcolor", "show_points", "pointsize"]
        if hasattr(self, "_lines_data"):
            options += ["linecolor", "show_lines", "linewidth"]
        if hasattr(self, "_frontfaces_data"):
            options += ["facecolor", "show_faces"]
        return options

    def make_buffer_from_data(self, data):
        """Create buffers from point/line/face data.

        Parameters
        ----------
        data: tuple
            Contains positions, colors, elements for the buffer

        Returns
        -------
        buffer_dict
           A dict with created buffer indexes
        """
        positions, colors, elements = data
        return {
            'positions': make_vertex_buffer(list(flatten(positions))),
            'colors': make_vertex_buffer(list(flatten(colors))),
            'elements': make_index_buffer(list(flatten(elements))),
            'n': len(list(flatten(elements)))
        }

    def update_buffer_from_data(self, data, buffer, update_positions=True, update_colors=True, update_elements=True):
        """Update existing buffers from point/line/face data.

        Parameters
        ----------
        data: tuple
            Contains positions, colors, elements for the buffer
        buffer: dict
            The dict with created buffer indexes
        update_positions : bool
            Whether to update positions in the buffer dict
        update_colors : bool
            Whether to update colors in the buffer dict
        update_elements : bool
            Whether to update elements in the buffer dict
        """
        positions, colors, elements = data
        if update_positions:
            update_vertex_buffer(list(flatten(positions)), buffer["positions"])
        if update_colors:
            update_vertex_buffer(list(flatten(colors)), buffer["colors"])
        if update_elements:
            update_index_buffer(list(flatten(elements)), buffer["elements"])
        buffer["n"] = len(list(flatten(elements)))

    def make_buffers(self):
        """Create all buffers from object's data"""
        if hasattr(self, '_points_data'):
            data = self._points_data()
            self._points_buffer = self.make_buffer_from_data(data)
            if data[0]:
                self._update_bounding_box(data[0])
        if hasattr(self, '_lines_data'):
            data = self._lines_data()
            self._lines_buffer = self.make_buffer_from_data(data)
            if data[0] and self._bounding_box_center is None:
                self._update_bounding_box(data[0])
        if hasattr(self, '_frontfaces_data'):
            data = self._frontfaces_data()
            self._frontfaces_buffer = self.make_buffer_from_data(data)
            if data[0] and self._bounding_box_center is None:
                self._update_bounding_box(data[0])
        if hasattr(self, '_backfaces_data'):
            data = self._backfaces_data()
            self._backfaces_buffer = self.make_buffer_from_data(data)
            if data[0] and self._bounding_box_center is None:
                self._update_bounding_box(data[0])

    def update_buffers(self):
        """Update all buffers from object's data"""
        if hasattr(self, '_points_data'):
            self.update_buffer_from_data(self._points_data(), self._points_buffer)
        if hasattr(self, '_lines_data'):
            self.update_buffer_from_data(self._lines_data(), self._lines_buffer)
        if hasattr(self, '_frontfaces_data'):
            self.update_buffer_from_data(self._frontfaces_data(), self._frontfaces_buffer)
        if hasattr(self, '_backfaces_data'):
            self.update_buffer_from_data(self._backfaces_data(), self._backfaces_buffer)

    def init(self):
        """Initialize the object"""
        self.make_buffers()
        self._update_matrix()

    def update(self):
        """Update the object"""
        self._update_matrix()
        self.update_buffers()

    def _update_bounding_box(self, positions=None):
        """Update the bounding box of the object"""
        if positions is None:
            positions = []
            if hasattr(self, '_points_data'):
                positions += self._points_data()[0]
            if hasattr(self, '_lines_data'):
                positions += self._lines_data()[0]
            if hasattr(self, '_frontfaces_data'):
                positions += self._frontfaces_data()[0]
            if not positions:
                return

        positions = np.array(positions)
        self._bounding_box = np.array([positions.min(axis=0), positions.max(axis=0)])
        self._bounding_box_center = np.average(self.bounding_box, axis=0)

    def draw(self, shader, wireframe=False, is_lighted=False):
        """Draw the object from its buffers"""
        shader.enable_attribute('position')
        shader.enable_attribute('color')
        shader.uniform1i('is_selected', self.is_selected)
        if self._matrix_buffer is not None:
            shader.uniform4x4('transform', self._matrix_buffer)
        shader.uniform1i('is_lighted', is_lighted)
        shader.uniform1f('object_opacity', self.opacity)
        shader.uniform1i('element_type', 2)
        if hasattr(self, "_frontfaces_buffer") and self.show_faces and not wireframe:
            shader.uniform3f('single_color', self.facecolor)
            shader.uniform1i('use_single_color', not self.facecolors and not self._is_collection and not getattr(self, 'use_vertex_color', False))
            shader.bind_attribute('position', self._frontfaces_buffer['positions'])
            shader.bind_attribute('color', self._frontfaces_buffer['colors'])
            shader.draw_triangles(elements=self._frontfaces_buffer['elements'], n=self._frontfaces_buffer['n'], background=self.background)
        if hasattr(self, "_backfaces_buffer") and self.show_faces and not wireframe:
            shader.uniform3f('single_color', self.facecolor)
            shader.uniform1i('use_single_color', not self.facecolors and not self._is_collection and not getattr(self, 'use_vertex_color', False))
            shader.bind_attribute('position', self._backfaces_buffer['positions'])
            shader.bind_attribute('color', self._backfaces_buffer['colors'])
            shader.draw_triangles(elements=self._backfaces_buffer['elements'], n=self._backfaces_buffer['n'], background=self.background)
        shader.uniform1i('is_lighted', False)
        shader.uniform1i('element_type', 1)
        if hasattr(self, "_lines_buffer") and (self.show_lines or wireframe):
            shader.uniform3f('single_color', self.linecolor)
            shader.uniform1i('use_single_color', not self.linecolors and not self._is_collection)
            shader.bind_attribute('position', self._lines_buffer['positions'])
            shader.bind_attribute('color', self._lines_buffer['colors'])
            shader.draw_lines(width=self.linewidth, elements=self._lines_buffer['elements'], n=self._lines_buffer['n'], background=self.background)
        shader.uniform1i('element_type', 0)
        if hasattr(self, "_points_buffer") and self.show_points:
            shader.uniform3f('single_color', self.pointcolor)
            shader.uniform1i('use_single_color', not self.pointcolors and not self._is_collection)
            shader.bind_attribute('position', self._points_buffer['positions'])
            shader.bind_attribute('color', self._points_buffer['colors'])
            shader.draw_points(size=self.pointsize, elements=self._points_buffer['elements'], n=self._points_buffer['n'], background=self.background)

        shader.uniform1i('is_selected', 0)
        shader.uniform1f('object_opacity', 1)
        if self._matrix_buffer is not None:
            shader.uniform4x4('transform', np.identity(4).flatten())
        shader.disable_attribute('position')
        shader.disable_attribute('color')

    def draw_instance(self, shader, wireframe=False):
        """Draw the object instance for picking"""
        shader.enable_attribute('position')
        shader.uniform3f('instance_color', self._instance_color)
        if self._matrix_buffer is not None:
            shader.uniform4x4('transform', self._matrix_buffer)
        if hasattr(self, "_points_buffer") and self.show_points:
            shader.bind_attribute('position', self._points_buffer['positions'])
            shader.draw_points(size=self.pointsize, elements=self._points_buffer['elements'], n=self._points_buffer['n'])
        if hasattr(self, "_lines_buffer") and (self.show_lines or wireframe):
            shader.bind_attribute('position', self._lines_buffer['positions'])
            shader.draw_lines(width=self.linewidth, elements=self._lines_buffer['elements'], n=self._lines_buffer['n'])
        if hasattr(self, "_frontfaces_buffer") and self.show_faces and not wireframe:
            shader.bind_attribute('position', self._frontfaces_buffer['positions'])
            shader.draw_triangles(elements=self._frontfaces_buffer['elements'], n=self._frontfaces_buffer['n'])
            shader.bind_attribute('position', self._backfaces_buffer['positions'])
            shader.draw_triangles(elements=self._backfaces_buffer['elements'], n=self._backfaces_buffer['n'])
        if self._matrix_buffer is not None:
            shader.uniform4x4('transform', np.identity(4).flatten())
        shader.uniform3f('instance_color', [0, 0, 0])
        shader.disable_attribute('position')



from compas.geometry import Vector
from compas.geometry import Shape


class Text(Shape):
    """
    """

    def __init__(
            self, text, position=[0, 0, 0], height=50):
        super().__init__()
        self.text = text
        self.position = Vector(*position)
        self.height = height



import math

from compas.geometry import Frame

from .polylineobject import PolylineObject


class EllipseObject(PolylineObject):
    """Object for displaying COMPAS Ellipse geometry."""

    def __init__(self, ellipse, u=16, **kwargs):
        self.u = u
        self.calculate_ellipse_points(ellipse)
        super().__init__(ellipse, close=True, **kwargs)

    def calculate_ellipse_points(self, ellipse):
        frame = Frame.from_plane(ellipse.plane)
        ellipse.points = [
            frame.to_world_coordinates([
                math.cos(i*math.pi*2/self.u) * ellipse.major,
                math.sin(i*math.pi*2/self.u) * ellipse.minor,
                0
            ])
            for i in range(self.u)
        ]

    def update(self):
        self.calculate_ellipse_points(self._data)
        self.init()
        super().update()

    @property
    def properties(self):
        return ["u"]



import numpy as np

from compas_view2.collections import Collection
from .object import Object
from .bufferobject import BufferObject


class CollectionObject(BufferObject):
    """Object for displaying COMPAS collection."""

    def __init__(self, collection: Collection, **kwargs):

        super().__init__(collection, **kwargs)

        self._objects = []
        for item, item_property in zip(collection.items, collection.item_properties):
            _kwargs = dict(kwargs)
            _kwargs.update(item_property)
            self._objects.append(Object.build(item, **_kwargs))

        # if not given explicitly, use child object default settings
        if self._objects:
            self.show_points = self.show_points or self._objects[0].show_points
            self.show_lines = self.show_lines or self._objects[0].show_lines
            self.show_faces = self.show_faces or self._objects[0].show_faces
        self._is_collection = True

    def _points_data(self):
        positions = []
        colors = []
        elements = []
        for obj in self._objects:
            if hasattr(obj, "_points_data"):
                p, c, e = obj._points_data()
                e = (np.array(e) + len(positions)).tolist()
                positions += p
                colors += c
                elements += e
        return positions, colors, elements

    def _lines_data(self):
        positions = []
        colors = []
        elements = []
        for obj in self._objects:
            if hasattr(obj, "_lines_data"):
                p, c, e = obj._lines_data()
                e = (np.array(e) + len(positions)).tolist()
                positions += p
                colors += c
                elements += e
        return positions, colors, elements

    def _frontfaces_data(self):
        positions = []
        colors = []
        elements = []
        for obj in self._objects:
            if hasattr(obj, "_frontfaces_data"):
                p, c, e = obj._frontfaces_data()
                e = (np.array(e) + len(positions)).tolist()
                positions += p
                colors += c
                elements += e
        return positions, colors, elements

    def _backfaces_data(self):
        positions = []
        colors = []
        elements = []
        for obj in self._objects:
            if hasattr(obj, "_backfaces_data"):
                p, c, e = obj._backfaces_data()
                e = (np.array(e) + len(positions)).tolist()
                positions += p
                colors += c
                elements += e
        return positions, colors, elements



from compas_view2.collections import Collection

from .collectionobject import CollectionObject

import itertools
from compas.geometry import Line
from compas.geometry import Plane
from compas.geometry import Circle
from compas.geometry import Transformation
from compas.geometry import Translation
from compas.geometry import Rotation
from compas.robots import Geometry
from math import radians


class RobotObject(CollectionObject):
    """Object for displaying COMPAS RobotModel.

    Attributes
    ----------
    robot : :class:`compas.robots.RobotModel`
        The compas robot model.
    joints : dict
        Dictionary of joints and their corresponding objects.
    link_objs : list
        Dictionary of links and their corresponding objects.
    """

    def __init__(self, robot, **kwargs):
        kwargs.update({'name': robot.name})
        super().__init__(Collection(), **kwargs)
        self.robot = robot
        self.joints = {}
        self.link_objs = {}
        self.create(robot.root, self)

    def create(self, link, parent, parent_joint=None):
        """Recursively create the robot joints and links in hierrachy.

        Parameters
        ----------
        link: :class:`compas.robots.Joint`
            The robot link to create.
        parent: :class:`compas_view2.objects.collectionobject.CollectionObject`
            The parent compas_view2 collection object.
        parent_joint: :class:`compas.robots.Joint`
            The parent joint of the link.

        Returns
        -------
        None

        """
        meshes = []

        for item in itertools.chain(link.visual):
            meshes.extend(Geometry._get_item_meshes(item))

        if meshes:
            obj = parent.add(Collection(meshes), name=link.name, show_lines=False)
        else:
            lines = []
            for joint in link.joints:
                line = Line([0, 0, 0], joint.origin.point)
                lines.append(line)
                if parent_joint:
                    lines.append(Circle(Plane([0, 0, 0], parent_joint.axis.vector), line.length/3))
            obj = parent.add(Collection(lines), name=link.name)

        if parent_joint:
            obj.matrix = Transformation.from_frame(parent_joint.origin).matrix
            self.joints[parent_joint.name] = {
                'link_obj': obj,
                'joint': parent_joint,
                'axis': parent_joint.axis.vector
            }

        self.link_objs[link.name] = obj

        for joint in link.joints:
            self.create(joint.child_link, parent=obj, parent_joint=joint)

    def rotate_joint(self, joint_name, angle):
        """Rotate the joint by the given angle.

        Parameters
        ----------
        joint_name: str
            The name of the joint to rotate.
        angle: float
            The angle to rotate the joint by.

        Returns
        -------
        None

        """
        joint = self.joints[joint_name]
        T = Translation.from_vector(joint['link_obj'].translation)
        R = Rotation.from_axis_and_angle(joint['axis'], radians(angle))
        joint['link_obj'].matrix = (T * R).matrix



from compas.geometry import Point

from .bufferobject import BufferObject


class PointObject(BufferObject):
    """Object for displaying COMPAS point geometry."""

    def __init__(self, data, **kwargs):
        super().__init__(data, **kwargs)
        self.show_points = True

    def _points_data(self):
        positions = [self._data]
        colors = [self.pointcolor]
        elements = [[0]]
        return positions, colors, elements

    @classmethod
    def create_default(cls) -> Point:
        return Point(0, 0, 0)



from compas.geometry import Frame
from compas.colors import Color
from .bufferobject import BufferObject


class FrameObject(BufferObject):
    """Object for displaying COMPAS Frame geometry."""

    def __init__(self, data, size=1.0, **kwargs):
        super().__init__(data, **kwargs)
        self.size = size
        self.show_points = True
        self.show_lines = True
        self.linecolors = [Color(1, 0, 0), Color(1, 0, 0), Color(0, 1, 0), Color(0, 1, 0), Color(0, 0, 1), Color(0, 0, 1)]

    def _points_data(self):
        frame = self._data
        positions = [frame.point]
        colors = [self.pointcolor]
        elements = [[0]]
        return positions, colors, elements

    def _lines_data(self):
        frame = self._data
        positions = [
            frame.point, frame.point + (frame.xaxis * self.size),
            frame.point, frame.point + (frame.yaxis * self.size),
            frame.point, frame.point + (frame.zaxis * self.size)]
        colors = self.linecolors
        elements = [[0, 1], [2, 3], [4, 5]]
        return positions, colors, elements

    @classmethod
    def create_default(cls) -> Frame:
        return Frame([0, 0, 0], [1, 0, 0], [0, 1, 0])



from OpenGL import GL

from qtpy import QtWidgets
from qtpy import QtCore

from typing import Union
from pathlib import Path

from compas.geometry import Point
from compas.datastructures import Network
from compas.datastructures import Mesh

from compas_view2.forms import AddForm
from compas_view2.forms import PropertyForm

from .worker import Worker


class Controller:
    """Action controller for the default config file.

    Parameters
    ----------
    app: :class:`compas_view2.app.App`
        The parent application.

    """

    def __init__(self, app):
        self.app = app

    def interactive(action="add"):
        """Decorator for transforming functions into "data add" or "object edit" actions.

        Parameters
        ----------
        action : {'add', 'edit'}, optional
            The type of action.

        Returns
        -------
        callable

        """
        def outer(func):
            def wrapped(self):
                def add(data):
                    if data:
                        self.app.add(data)
                        self.app.view.update()

                def edit(obj):
                    if obj:
                        dock = PropertyForm("Property", obj, on_update=self.app.view.update)
                        self.app.window.addDockWidget(QtCore.Qt.RightDockWidgetArea, dock)

                worker = Worker(func, [self])
                if action == "add":
                    worker.signals.result.connect(add)
                elif action == "edit":
                    worker.signals.result.connect(edit)
                else:
                    raise NotImplementedError()
                Worker.pool.start(worker)
            return wrapped
        return outer

    # ==============================================================================
    # App actions
    # ==============================================================================

    def dummy(self):
        """Dummy action for UI element stubs.

        Returns
        -------
        None

        """
        pass

    def about(self):
        """Display the about message.

        Returns
        -------
        None

        """
        self.app.about()

    def opengl_version(self):
        """Display the OpenGL version.

        Returns
        -------
        None

        """
        value = "OpenGL {}".format(GL.glGetString(GL.GL_VERSION).decode('ascii'))
        self.app.info(value)

    def glsl_version(self):
        """Display the version of the shader language.

        Returns
        -------
        None

        """
        value = "GLSL {}".format(GL.glGetString(GL.GL_SHADING_LANGUAGE_VERSION).decode('ascii'))
        self.app.info(value)

    # ==============================================================================
    # View actions
    # ==============================================================================

    def view_shaded(self):
        """Switch the view to shaded.

        Returns
        -------
        None

        """
        self.app.view.mode = 'shaded'
        self.app.view.update()

    def view_ghosted(self):
        """Switch the view to ghosted.

        Returns
        -------
        None

        """
        self.app.view.mode = 'ghosted'
        self.app.view.update()

    def view_wireframe(self):
        """Switch the view to wireframe.

        Returns
        -------
        None

        """
        self.app.view.mode = 'wireframe'
        self.app.view.update()

    def view_lighted(self):
        """Switch the view to lighted.

        Returns
        -------
        None

        """
        self.app.view.mode = 'lighted'
        self.app.view.update()

    def view_capture(self, filepath=None):
        """Capture a screenshot.

        Parameters
        ----------
        filepath : str, optional
            The destination path for saving the screenshot.
            If no path is provided, a file dialog will be be opened automatically.

        Returns
        -------
        None

        """
        if filepath:
            result = filepath
        else:
            result = QtWidgets.QFileDialog.getSaveFileName(caption="File name", dir="")
            if not result:
                return
            result = result[0]
        filepath = Path(result)
        if not filepath.suffix:
            return
        qimage = self.app.view.grabFramebuffer()
        qimage.save(str(filepath), filepath.suffix[1:])

    def view_front(self):
        """Swtich to a front view.

        Returns
        -------
        None

        """
        self.app.view.current = self.app.view.FRONT
        self.app.view.camera.reset_position()
        self.app.view.update_projection()
        self.app.view.update()

    def view_right(self):
        """Swtich to a right view.

        Returns
        -------
        None

        """
        self.app.view.current = self.app.view.RIGHT
        self.app.view.camera.reset_position()
        self.app.view.update_projection()
        self.app.view.update()

    def view_top(self):
        """Swtich to a top view.

        Returns
        -------
        None

        """
        self.app.view.current = self.app.view.TOP
        self.app.view.camera.reset_position()
        self.app.view.update_projection()
        self.app.view.update()

    def view_perspective(self):
        """Swtich to a perspective view.

        Returns
        -------
        None

        """
        self.app.view.current = self.app.view.PERSPECTIVE
        self.app.view.camera.reset_position()
        self.app.view.update_projection()
        self.app.view.update()

    # ==============================================================================
    # Scene actions
    # ==============================================================================

    def load_scene(self):
        """Load a scene from a file."""
        self.app.info('Not available yet...')

    def save_scene(self):
        """Save the current scene to a file."""
        self.app.info('Not available yet...')

    def redraw_scene(self):
        """Redraw the current scene.

        Returns
        -------
        None

        """
        self.app.view.update()

    def clear_scene(self):
        """Clear all objects from the current scene."""
        self.app.info('Not available yet...')

    def undo(self):
        """Undo the last scene modification."""
        self.app.info('Not available yet...')

    def redo(self):
        """Redo the last scene modification."""
        self.app.info('Not available yet...')

    def history(self):
        """Display the undo history of the scene."""
        self.app.info('Not available yet...')

    # ==============================================================================
    # Primitive actions
    # ==============================================================================

    def add_geometry_object(self):
        def on_create(data):
            self.app.add(data)
            self.app.view.update()
        form = AddForm(on_create=on_create)
        form.exec_()

    @interactive("edit")
    def edit_selected_object(self):
        if len(self.app.selector.selected) == 1:
            return self.app.selector.selected[0]
        else:
            self.app.status("Select one object on screen, Click Enter to finish")
            objects = self.app.selector.start_selection(types=[Point], mode="single", returns="object")
            if len(objects) != 1:
                self.app.status("Must select 1 object")
                return None
            self.app.status("")
            return objects[0]

    # ==============================================================================
    # Network actions
    # ==============================================================================

    def add_network_from_obj(self) -> Union[Network, None]:
        """Add a network from the data in an OBJ file.

        Returns
        -------
        :class:`compas.datastructures.Network` or None
            A network data structure,
            if the operation was successful,
            or None otherwise.

        """
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(parent=self.app.window,
                                                            caption="Select File",
                                                            dir="",
                                                            filter="OBJ Files (*.obj)")
        if filename:
            network = Network.from_obj(filename)
            self.app.add(network)
            self.app.view.update()
            return network

    # ==============================================================================
    # Mesh actions
    # ==============================================================================

    def add_mesh_from_file(self) -> Union[Mesh, None]:
        """Add a mesh from the data in an file.

        Returns
        -------
        :class:`compas.datastructures.Mesh` or None
            A mesh data structure,
            if the operation was successful,
            or None otherwise.

        Notes
        -----
        The following file formats are supported

        * OBJ
        * OFF
        * PLY
        * STL

        """
        filepath, selectedfilter = QtWidgets.QFileDialog.getOpenFileName(
            parent=self.app.window,
            caption="Select File",
            dir="",
            filter="Mesh Files (*.obj, *.off, *.ply, *.stl)")

        if filepath:
            path = Path(filepath)

            if path.suffix == '.obj':
                mesh = Mesh.from_obj(path)
            elif path.suffix == '.off':
                mesh = Mesh.from_off(path)
            elif path.suffix == '.ply':
                mesh = Mesh.from_ply(path)
            elif path.suffix == '.stl':
                mesh = Mesh.from_stl(path)

            self.app.add(mesh)
            self.app.view.update()
            return mesh

    # ==============================================================================
    # Flow actions
    # ==============================================================================

    def show_flow(self) -> None:
        """Display the ryven flow window."""
        self.app.flow.show()

    def run_all(self) -> None:
        """Execute all the ryven nodes in the order of data flow."""
        self.app.flow.run_all()

    def enable_auto_update(self) -> None:
        """Enable auto update of all the ryven nodes."""
        self.run_all()
        for node in self.app.flow.flow_view.node_items:
            node.auto_update = True

    def disable_auto_update(self) -> None:
        """Disable auto update of all the ryven nodes."""
        for node in self.app.flow.flow_view.node_items:
            node.auto_update = False

    # ==============================================================================
    # Robot actions
    # ==============================================================================

    # ==============================================================================
    # Assembly actions
    # ==============================================================================

    # ==============================================================================
    # FEA actions
    # ==============================================================================



from compas.utilities import flatten
from .bufferobject import BufferObject


class NurbsSurfaceObject(BufferObject):
    """Object for displaying COMPAS NurbsSurface geometry."""

    def __init__(self, surface, u=100, v=100, **kwargs):
        super().__init__(surface, **kwargs)
        self._data = surface
        self._triangles = [list(point) for triangle in surface.to_triangles(nu=u, nv=v) for point in triangle]
        self.u = u
        self.v = v

    def update(self):
        self._triangles = [list(point) for triangle in self._data.to_triangles(nu=self.u, nv=self.v) for point in triangle]
        self.init()
        super().update()

    def _points_data(self):
        positions = [list(pt) for pt in flatten(self._data.points)]
        colors = [self.pointcolor for _ in positions]
        elements = [[i] for i in range(len(positions))]
        return positions, colors, elements

    def _lines_data(self):
        positions = [list(pt) for pt in flatten(self._data.points)]
        colors = [self.linecolor for _ in positions]
        count = 0
        indexes = []
        for row in self._data.points:
            row_i = []
            for _ in range(len(row)):
                row_i.append(count)
                count += 1
            indexes.append(row_i)
        elements = []
        for row in indexes:
            for i in range(len(row) - 1):
                elements.append([row[i], row[i + 1]])
        for col in zip(*indexes):
            for i in range(len(col) - 1):
                elements.append([col[i], col[i + 1]])
        return positions, colors, elements

    def _frontfaces_data(self):
        positions = self._triangles
        colors = [self.facecolor] * len(self._triangles)
        elements = [[i * 3 + 0, i * 3 + 1, i * 3 + 2] for i in range(int(len(self._triangles) / 3))]
        return positions, colors, elements

    def _backfaces_data(self):
        positions = self._triangles[::-1]
        colors = [self.facecolor] * len(self._triangles)
        elements = [[i * 3 + 0, i * 3 + 1, i * 3 + 2] for i in range(int(len(self._triangles) / 3))]
        return positions, colors, elements

    @property
    def properties(self):
        return ["u", "v"]



from compas.data import Data
from typing import Tuple, Union, List


class Collection:
    """A collection of COMPAS items like meshes or shapes.
    """

    def __init__(self, items: Union[Data, Tuple[Data, dict]] = None, item_properties: List[dict] = None):
        super().__init__()
        self.items = []
        self.item_properties = []
        items = items or []
        item_properties = item_properties or []
        for item in items:
            if isinstance(item, Data):
                self.item_properties.append({})
            else:
                item, properties = item
                self.item_properties.append(properties)
            self.items.append(item)

        if item_properties:
            self.item_properties = item_properties

        assert len(self.items) == len(self.item_properties)



from compas.geometry import centroid_points
from compas.utilities import pairwise
from .polylineobject import PolylineObject


class PolygonObject(PolylineObject):
    """Object for displaying COMPAS Polygon geometry."""

    def __init__(self, polygon, show_face=True, facecolor=None, **kwargs):
        super().__init__(polygon, close=True, facecolor=facecolor, **kwargs)
        self.show_face = show_face

    def _frontfaces_data(self):
        if not self.show_face:
            return
        positions = []
        colors = []
        elements = []
        points = self._data.points
        color = self.facecolor
        if len(points) == 3:
            a, b, c = points
            positions.append(a)
            positions.append(b)
            positions.append(c)
            colors.append(color)
            colors.append(color)
            colors.append(color)
            elements.append([0, 1, 2])
        elif len(points) == 4:
            a, b, c, d = points
            positions.append(a)
            positions.append(b)
            positions.append(c)
            positions.append(a)
            positions.append(c)
            positions.append(d)
            colors.append(color)
            colors.append(color)
            colors.append(color)
            colors.append(color)
            colors.append(color)
            colors.append(color)
            elements.append([0, 1, 2])
            elements.append([3, 4, 5])
        else:
            c = centroid_points(points)
            i = 0
            for a, b in pairwise(points + points[:1]):
                positions.append(a)
                positions.append(b)
                positions.append(c)
                colors.append(color)
                colors.append(color)
                colors.append(color)
                elements.append([i + 0, i + 1, i + 2])
                i += 3
        return positions, colors, elements

    def _backfaces_data(self):
        if not self.show_face:
            return
        positions = []
        colors = []
        elements = []
        points = self._data.points[::-1]
        color = self.facecolor
        if len(points) == 3:
            a, b, c = points
            positions.append(a)
            positions.append(b)
            positions.append(c)
            colors.append(color)
            colors.append(color)
            colors.append(color)
            elements.append([0, 1, 2])
        elif len(points) == 4:
            a, b, c, d = points
            positions.append(a)
            positions.append(b)
            positions.append(c)
            positions.append(a)
            positions.append(c)
            positions.append(d)
            colors.append(color)
            colors.append(color)
            colors.append(color)
            colors.append(color)
            colors.append(color)
            colors.append(color)
            elements.append([0, 1, 2])
            elements.append([3, 4, 5])
        else:
            c = centroid_points(points)
            i = 0
            for a, b in pairwise(points + points[:1]):
                positions.append(a)
                positions.append(b)
                positions.append(c)
                colors.append(color)
                colors.append(color)
                colors.append(color)
                elements.append([i + 0, i + 1, i + 2])
                i += 3
        return positions, colors, elements



import ryvencore_qt as rc
import inspect
from compas_view2.objects import DATA_OBJECT
from qtpy.QtGui import QColor
from .widgets import ExecutionControl
from compas.colors import Color
from typing import Union
import traceback


def Node(app, color: Union[Color, str, list, tuple] = '#0092D2', auto_update: bool = None, **kwargs):
    """Decorator for creating a custom ryven node in compas_view2 flow.

        Parameters
        ----------
        app: :class:`compas_view2.app.App`
            The compas_view2 app instance.
        color : Union[Color, str, list, tuple]
            The color theme of the ndoe.
            Defaults to '#0092D2'.
        auto_update : bool
            Whether to auto update the node.
            Defaults to the global flow setting.
        **kwargs : dict
            Additional keyword arguments for the node output visualisation.

        Returns
        -------
        callable
            A CustomNode class that wraps the decorated function.

    """

    if Color.is_hex(color):
        color = Color.from_hex(color)
    elif Color.is_rgb255(color):
        color = Color.from_rgb255(*color)
    elif Color.is_rgb1(color):
        color = Color(*color)
    else:
        raise ValueError('Invalid color: {}'.format(color))

    node_color = color.hex

    def decorator(func):

        signature = inspect.signature(func)

        if auto_update is None:
            _auto_update = app.flow.flow_auto_update
        else:
            _auto_update = auto_update

        class CustomNode(rc.Node):
            """Class that wraps the decorated function."""

            title = func.__name__
            init_inputs = [rc.NodeInputBP(label=name) for name in signature.parameters.keys() if name != 'self']
            init_outputs = [rc.NodeOutputBP(signature.return_annotation.__name__)]
            color = node_color
            main_widget_class = ExecutionControl
            main_widget_pos = 'below ports'  # or 'between ports'

            def __init__(self, params):
                super().__init__(params)
                self.block_updates = not _auto_update
                self.actions['execute'] = {'method': self.update_event}
                self.actions['show_object'] = {'method': self.show_object}
                self.actions['hide_object'] = {'method': self.hide_object}
                self.actions['select_object'] = {'method': self.select_object}
                self.object = None
                self.app = app
                self.object_properties = kwargs
                if not self.object_properties.get('facecolor'):
                    self.object_properties['facecolor'] = color

            def __repr__(self) -> str:
                return f'<{self.__class__.__name__}({self.title})>'

            @property
            def auto_update(self):
                return not self.block_updates

            @auto_update.setter
            def auto_update(self, value):
                """Two direction binding to the checkbox widget."""
                self.block_updates = not value
                self.item.main_widget.set_auto_update(None, value=value, update_node=False)

            def select_object(self):
                """Select the object in the scene."""
                if self.object:
                    self.app.selector.select(self.object)
                    self.app.view.update()

            def show_object(self):
                """Show the object in the scene."""
                if self.object:
                    self.object_properties['is_visible'] = self.object.is_visible = True
                    self.object.is_visible = True
                    self.app.view.update()

            def hide_object(self):
                """Hide the object in the scene."""
                if self.object:
                    self.object_properties['is_visible'] = self.object.is_visible = False
                    self.app.view.update()

            def place_event(self):
                """Called upon node placement."""
                # This is to suppress a ryven exception to parse and empty dict when initiating the main widget
                self.init_data = None

            def remove_event(self):
                """Called upon node removal."""
                if self.object:
                    app.remove(self.object)
                    app.view.update()

            def update_event(self, inp=-1):
                """execute wrapped function."""

                if not self.app.started:
                    return

                _inputs = [self.input(i) for i in range(len(self.init_inputs))]
                try:
                    _output = func(*_inputs)
                    # Restore to default color if the function succeeded
                    self.item.main_widget.set_message()
                    self.change_color()

                except Exception as e:
                    _output = None
                    # Change node color and display error message
                    print("Function failed at", self)
                    print(traceback.format_exc())
                    self.item.main_widget.set_message(str(e))
                    self.change_color('#FF0000')

                finally:
                    if self.object:
                        app.remove(self.object)

                    if _output and _output.__class__ in DATA_OBJECT:
                        self.object = app.add(_output, **self.object_properties)

                    app.view.update()
                    self.set_output_val(0, _output)

            def change_color(self, color=None):
                """Change the theme color of the node."""
                color = color or self.color
                self.item.animator.stop()  # The animator must be stopped before changing the color
                self.item.color = QColor(color)
                self.item.update_design()

        app.flow.session.register_node(CustomNode)
        return CustomNode

    return decorator



from compas.datastructures import Mesh
from .meshobject import MeshObject


class PolyhedronObject(MeshObject):
    """Object for displaying COMPAS Polyhedron geometry."""

    def __init__(self, data, **kwargs):
        super().__init__(Mesh.from_shape(data), **kwargs)
        self._data = data



from compas.utilities import pairwise
from .bufferobject import BufferObject


class PolylineObject(BufferObject):
    """Object for displaying COMPAS Polyline geometry."""

    def __init__(self, data, close=False, **kwargs):
        super().__init__(data, **kwargs)
        self.close = close
        self.show_lines = True

    def _points_data(self):
        polyline = self._data
        positions = [point for point in polyline.points]
        colors = [self.pointcolors.get(i, self.pointcolor) for i, _ in enumerate(polyline.points)]
        elements = [[i] for i in range(len(positions))]
        return positions, colors, elements

    def _lines_data(self):
        polyline = self._data
        positions = []
        colors = []
        elements = []
        if self.close:
            lines = pairwise(polyline.points + [polyline.points[0]])
        else:
            lines = pairwise(polyline.points)
        count = 0
        for i, (pt1, pt2) in enumerate(lines):
            positions.append(pt1)
            positions.append(pt2)
            color = self.linecolors.get(i, self.linecolor)
            colors.append(color)
            colors.append(color)
            elements.append([count, count+1])
            count += 2
        return positions, colors, elements



import ryvencore_qt as rc
from typing import List, Tuple, Union
from compas.geometry import Vector as Vector


class ValueNode(rc.Node):
    """Base class for all input value nodes."""

    title = 'Value'
    color = '#0092D2'
    init_inputs = [rc.NodeInputBP()]
    init_outputs = [rc.NodeOutputBP()]

    @property
    def auto_update(self):
        return not self.block_updates

    @auto_update.setter
    def auto_update(self, value: bool):
        # value nodes should always auto update
        pass

    def view_place_event(self):
        self.update()

    def update_event(self, inp=-1):
        self.set_output_val(0, self.input(0))


def IntegerNode(title: str = 'Integer', default: int = 1, bounds: Tuple[float, float] = None) -> ValueNode:
    """Builds an integer node class.

        Parameters
        ----------
        title : str
            Displayed title of node.
            Defaults to 'Integer'.
        default : int
            Default value.
            Defaults to 1.
        bounds : Tuple[float, float]
            Min and max bounds of value.
            Defaults to (-10**9, 10**9).

        Returns
        -------
        :class:`compas_view2.flow.ValueNode`
            An extension of :class:`compas_view2.flow.ValueNode` that represents an integer value.

    """
    _title = title
    bounds = bounds or (-10**9, 10**9)

    class IntegerNode(ValueNode):

        title = _title

        init_inputs = [
            rc.NodeInputBP(
                dtype=rc.dtypes.Integer(default=default, bounds=bounds),
                label='Int'
            ),
        ]

    return IntegerNode


def FloatNode(title: str = 'Float', default: float = 0.0, bounds: tuple = None) -> ValueNode:
    """Builds an float node class.

        Parameters
        ----------
        title : str
            Displayed title of node.
            Defaults to 'Float'.
        default : float
            Default value.
            Defaults to 0.0.
        bounds : Tuple[float, float]
            Min and max bounds of value.
            Defaults to (-10**9, 10**9).

        Returns
        -------
        :class:`compas_view2.flow.ValueNode`
            An extension of :class:`compas_view2.flow.ValueNode` that represents an float value.

    """

    _title = title
    bounds = bounds or (-10**9, 10**9)

    class FloatNode(ValueNode):

        title = _title

        init_inputs = [
            rc.NodeInputBP(
                dtype=rc.dtypes.Float(default=default, bounds=bounds),
                label='Float'
            ),
        ]

    return FloatNode


def ChoiceNode(title: str = 'Choice', default: str = None, items: List[str] = None) -> ValueNode:
    """Builds an choice node class.

        Parameters
        ----------
        title : str
            Displayed title of node.
            Defaults to 'Choice'.
        default : str
            Default choice.
            Defaults to None.
        items : List[str]
            List of items to choose from.
            Defaults to [].

        Returns
        -------
        :class:`compas_view2.flow.ValueNode`
            An extension of :class:`compas_view2.flow.ValueNode` that represents a choice from list of values.

    """

    _title = title
    items = items or []

    class ChoiceNode(ValueNode):

        title = _title

        init_inputs = [
            rc.NodeInputBP(
                dtype=rc.dtypes.Choice(default=default, items=items),
                label='Choice'
            ),
        ]

    return ChoiceNode


def StringNode(title: str = 'String', default: str = None) -> ValueNode:
    """Builds an string node class.

        Parameters
        ----------
        title : str
            Displayed title of node.
            Defaults to 'String'.
        default : str
            Default string.
            Defaults to None.

        Returns
        -------
        :class:`compas_view2.flow.ValueNode`
            An extension of :class:`compas_view2.flow.ValueNode` that represents a string value.

    """

    _title = title

    class StringNode(ValueNode):

        title = _title

        init_inputs = [
            rc.NodeInputBP(
                dtype=rc.dtypes.String(default=default, size='l'),
                label='String'
            ),
        ]

    return StringNode


def VectorNode(title: str = 'Vector', default: Union[List[float], Tuple[float, float, float], Vector] = None) -> ValueNode:
    """Builds an vector node class.

        Parameters
        ----------
        title : str
            Displayed title of node.
            Defaults to 'Vector'.
        default : Union[List[float], Tuple[float, float, float], Vector]
            Default vector.
            Defaults to [0, 0, 0].

        Returns
        -------
        :class:`compas_view2.flow.ValueNode`
            An extension of :class:`compas_view2.flow.ValueNode` that represents a vector value.

    """

    _title = title
    default = default or [0, 0, 0]

    class VectorNode(ValueNode):

        title = _title

        init_inputs = [
            rc.NodeInputBP(
                dtype=rc.dtypes.Float(default=default[0]),
                label='x'
            ),
            rc.NodeInputBP(
                dtype=rc.dtypes.Float(default=default[1]),
                label='y'
            ),
            rc.NodeInputBP(
                dtype=rc.dtypes.Float(default=default[2]),
                label='z'
            ),
        ]

        def update_event(self, inp=-1):
            self.set_output_val(0, Vector(self.input(0), self.input(1), self.input(2)))

    return VectorNode



from qtpy import QtWidgets


class Button:
    """Class representing a button wrapped in a horizontal box layout.

    Parameters
    ----------
    app : :class:`compas_view2.app.App`
        The app containing the widget.
    parent : QtWidgets.QWidget
        The parent widget for the button.
    text : str
        The text label of the button.
    action : callable
        The action associated with the button.

    Attributes
    ----------
    action : callable
        Action associated with the button click event.
    button : QtWidgets.QPushButton
        The actual button widget.

    """

    def __init__(self,
                 app,
                 parent,
                 *,
                 text,
                 action):
        box = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        box.setLayout(layout)
        button = QtWidgets.QPushButton(text)
        layout.addWidget(button)
        parent.addWidget(box)
        # connect
        button.clicked.connect(action)
        button.clicked.connect(app.view.update)
        # attributes
        self.action = action
        self.button = button

    def __call__(self, *args, **kwargs):
        """Wrapper for the action associated with the checkbox.

        Returns
        -------
        None

        """
        return self.action()



from compas.geometry import Line
from .bufferobject import BufferObject


class LineObject(BufferObject):
    """Object for displaying COMPAS line geometry."""

    default_color_points = [0.1, 0.1, 0.1]
    default_color_lines = [0.4, 0.4, 0.4]

    def __init__(self, data, **kwargs):
        super().__init__(data, **kwargs)
        self.show_lines = True

    def _points_data(self):
        line = self._data
        color = self.pointcolor
        positions = [line.start, line.end]
        colors = [color, color]
        elements = [[0], [1]]
        return positions, colors, elements

    def _lines_data(self):
        line = self._data
        color = self.linecolor
        positions = [line.start, line.end]
        colors = [color, color]
        elements = [[0, 1]]
        return positions, colors, elements

    @classmethod
    def create_default(cls) -> Line:
        return Line([0, 0, 0], [0, 0, 1])



from compas_view2.shaders import Shader
from .view import View


class View330(View):
    """View widget for OpenGL 3.3 and GLSL 330 and above, with a Core Profile.
    """

    def init(self):
        # init the buffers
        for guid in self.objects:
            obj = self.objects[guid]
            obj.init()
        # create the program
        self.shader = Shader()
        self.shader.bind()
        self.shader.uniform4x4(
            "projection", self.camera.projection(self.app.width, self.app.height))
        self.shader.uniform4x4("viewworld", self.camera.viewworld())
        self.shader.uniform1i("is_selected", 0)
        self.shader.uniform1f("opacity", self.opacity)
        self.shader.uniform3f("selection_color", self.selection_color)
        self.shader.release()

    def resize(self, w: int, h: int):
        self.shader.bind()
        self.shader.uniform4x4("projection", self.camera.projection(w, h))
        self.shader.release()

    def paint(self):
        self.shader.bind()
        self.shader.uniform4x4("viewworld", self.camera.viewworld())
        for guid in self.objects:
            obj = self.objects[guid]
            obj.draw(self.shader)
        self.shader.release()



from compas.datastructures import Mesh
from compas.geometry import Box
from compas.geometry import Frame

from .meshobject import MeshObject


class BoxObject(MeshObject):
    """Object for displaying COMPAS box geometry."""

    def __init__(self, data, **kwargs):
        super().__init__(Mesh.from_shape(data), **kwargs)
        self._data = data

    def update(self):
        self._mesh = Mesh.from_shape(self._data)
        super().update()

    @classmethod
    def create_default(cls) -> Box:
        return Box(Frame.worldXY(), 1.0, 1.0, 1.0)



from qtpy import QtWidgets
from qtpy import QtGui


class Form(QtWidgets.QDialog):

    def __init__(self, title):
        super().__init__(f=QtGui.Qt.WindowTitleHint | QtGui.Qt.WindowSystemMenuHint)
        self.setWindowTitle(title)
        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        # inputs
        self._inputs = self.inputs()
        # buttons
        buttons = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addLayout(self._inputs)
        layout.addWidget(buttons)

    def inputs(self):
        raise NotImplementedError



import os
from OpenGL import GL


class Shader:
    """The shader used by the OpenGL view."""

    def __init__(self, name='120/mesh'):
        self.program = make_shader_program(name)
        self.locations = {}

    def uniform4x4(self, name, value):
        """Store a uniform 4x4 transformation matrix in the shader program at a named location.

        Parameters
        ----------
        name: str
            The name of the location in the shader program.
        value: array-like
            A 4x4 transformation matrix in column-major ordering.
        """
        loc = GL.glGetUniformLocation(self.program, name)
        GL.glUniformMatrix4fv(loc, 1, True, value)

    def uniform1i(self, name, value):
        """Store a uniform integer in the shader program at a named location.

        Parameters
        ----------
        name: str
            The name of the location in the shader program.
        value: int
            An integer value.
        """
        loc = GL.glGetUniformLocation(self.program, name)
        GL.glUniform1i(loc, value)

    def uniform1f(self, name, value):
        """Store a uniform float in the shader program at a named location.

        Parameters
        ----------
        name: str
            The name of the location in the shader program.
        value: float
            A float value.
        """
        loc = GL.glGetUniformLocation(self.program, name)
        GL.glUniform1f(loc, value)

    def uniform3f(self, name, value):
        """Store a uniform list of 3 floats in the shader program at a named location.

        Parameters
        ----------
        name: str
            The name of the location in the shader program.
        value: (float, float, float) | list[float]
            An iterable of 3 floats.
        """
        loc = GL.glGetUniformLocation(self.program, name)
        GL.glUniform3f(loc, *value)

    def uniformTex(self, name, texture):
        # loc = GL.glGetUniformLocation(self.program, name)
        # print(loc)
        GL.glActiveTexture(GL.GL_TEXTURE0 + 0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, texture)

    def bind(self):
        """Bind the shader program."""
        GL.glUseProgram(self.program)

    def release(self):
        """Release (unbind) the shader program."""
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, 0)
        GL.glUseProgram(0)

    def enable_attribute(self, name):
        location = GL.glGetAttribLocation(self.program, name)
        GL.glEnableVertexAttribArray(location)
        self.locations[name] = location

    def bind_attribute(self, name, value, step=3):
        location = self.locations[name]
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, value)
        GL.glVertexAttribPointer(location, step, GL.GL_FLOAT, False, 0, None)

    def disable_attribute(self, name):
        GL.glDisableVertexAttribArray(self.locations[name])
        del self.locations[name]

    def draw_triangles(self, elements=None, n=0, background=False):
        if elements:
            if background:
                GL.glDisable(GL.GL_DEPTH_TEST)
            GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, elements)
            GL.glDrawElements(GL.GL_TRIANGLES, n, GL.GL_UNSIGNED_INT, None)
        else:
            GL.glDrawArrays(GL.GL_TRIANGLES, 0, GL.GL_BUFFER_SIZE)

    def draw_lines(self, elements=None, n=0, width=1, background=False):
        if elements:
            if background:
                GL.glDisable(GL.GL_DEPTH_TEST)
            GL.glLineWidth(width)
            GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, elements)
            GL.glDrawElements(GL.GL_LINES, n, GL.GL_UNSIGNED_INT, None)
            GL.glEnable(GL.GL_DEPTH_TEST)
        else:
            GL.glDrawArrays(GL.GL_LINES, 0, GL.GL_BUFFER_SIZE)

    def draw_points(self, size=1, elements=None, n=0, background=False):
        GL.glPointSize(size)
        if elements:
            if background:
                GL.glDisable(GL.GL_DEPTH_TEST)
            GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, elements)
            GL.glDrawElements(GL.GL_POINTS, n, GL.GL_UNSIGNED_INT, None)
        else:
            GL.glDrawArrays(GL.GL_POINTS, 0, GL.GL_BUFFER_SIZE)

    def draw_texts(self, elements=None, n=0):
        GL.glDisable(GL.GL_POINT_SMOOTH)
        GL.glEnable(GL.GL_POINT_SPRITE)
        GL.glEnable(GL.GL_PROGRAM_POINT_SIZE)
        if elements:
            GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, elements)
            GL.glDrawElements(GL.GL_POINTS, n, GL.GL_UNSIGNED_INT, None)
        else:
            GL.glDrawArrays(GL.GL_POINTS, 0, GL.GL_BUFFER_SIZE)
        GL.glDisable(GL.GL_POINT_SPRITE)
        GL.glEnable(GL.GL_POINT_SMOOTH)

    def draw_arrows(self, elements=None, n=0):
        GL.glDisable(GL.GL_POINT_SMOOTH)
        GL.glEnable(GL.GL_POINT_SPRITE)
        GL.glEnable(GL.GL_PROGRAM_POINT_SIZE)
        if elements:
            GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, elements)
            GL.glDrawElements(GL.GL_POINTS, n, GL.GL_UNSIGNED_INT, None)
        else:
            GL.glDrawArrays(GL.GL_POINTS, 0, GL.GL_BUFFER_SIZE)
        GL.glDisable(GL.GL_POINT_SPRITE)
        GL.glEnable(GL.GL_POINT_SMOOTH)

    def draw_2d_box(self, box_coords, width, height):
        x1, y1, x2, y2 = box_coords
        x1 = (x1/width - 0.5)*2
        x2 = (x2/width - 0.5)*2
        y1 = -(y1/height - 0.5)*2
        y2 = -(y2/height - 0.5)*2
        if x1 > x2:
            x1, x2 = x2, x1
        if y1 > y2:
            y1, y2 = y2, y1
        GL.glLineWidth(1)
        GL.glBegin(GL.GL_LINE_LOOP)
        GL.glColor3f(0, 0, 0)
        GL.glVertex2f(x1, y1)
        GL.glVertex2f(x2, y1)
        GL.glVertex2f(x2, y2)
        GL.glVertex2f(x1, y2)
        GL.glEnd()


def make_shader_program(name):
    vsource = os.path.join(os.path.dirname(__file__), "{}.vert".format(name))
    fsource = os.path.join(os.path.dirname(__file__), "{}.frag".format(name))

    with open(vsource, "r") as f:
        vertex = compile_vertex_shader(f.read())

    with open(fsource, "r") as f:
        fragment = compile_fragment_shader(f.read())

    program = GL.glCreateProgram()
    GL.glAttachShader(program, vertex)
    GL.glAttachShader(program, fragment)
    GL.glLinkProgram(program)
    GL.glValidateProgram(program)
    result = GL.glGetProgramiv(program, GL.GL_LINK_STATUS)
    if not result:
        raise RuntimeError(GL.glGetProgramInfoLog(program))
    GL.glDeleteShader(vertex)
    GL.glDeleteShader(fragment)
    return program


def compile_vertex_shader(source):
    shader = GL.glCreateShader(GL.GL_VERTEX_SHADER)
    GL.glShaderSource(shader, source)
    GL.glCompileShader(shader)
    result = GL.glGetShaderiv(shader, GL.GL_COMPILE_STATUS)
    if not result:
        raise RuntimeError(GL.glGetShaderInfoLog(shader))
    return shader


def compile_fragment_shader(source):
    shader = GL.glCreateShader(GL.GL_FRAGMENT_SHADER)
    GL.glShaderSource(shader, source)
    GL.glCompileShader(shader)
    result = GL.glGetShaderiv(shader, GL.GL_COMPILE_STATUS)
    if not result:
        raise RuntimeError(GL.glGetShaderInfoLog(shader))
    return shader



from .dockform import DockForm
from qtpy import QtWidgets
from qtpy import QtCore


class SceneForm(DockForm):
    def __init__(self, app, title="Scene"):
        super().__init__(app, title)
        self.tree = QtWidgets.QTreeWidget()
        self.tree.setColumnCount(1)
        self.tree.setHeaderHidden(True)
        self.setWidget(self.tree)
        self.tree.itemPressed.connect(lambda item: self.show_properties(item.obj))

    def update(self):
        self.tree.clear()
        root_objects = filter(lambda obj: obj.parent is None,  self.app.view.objects.values())
        self.map_objects(root_objects)

    def map_objects(self, objects, parent=None):
        for obj in objects:
            item = QtWidgets.QTreeWidgetItem([obj.name])
            if parent is not None:
                parent.addChild(item)
            else:
                self.tree.addTopLevelItem(item)

            item.obj = obj

            self.map_objects(obj.children, item)

    def show_properties(self, obj):

        self.app.selector.select(obj, update=True)

        propertyform = self.app.dock_slots.get("propertyform")
        if propertyform:
            propertyform.set_object(obj)

    def select(self, objs):
        all_items = self.tree.findItems("", QtCore.Qt.MatchContains | QtCore.Qt.MatchRecursive)
        for item in all_items:
            if item.obj in objs:
                item.setSelected(True)
            else:
                item.setSelected(False)



from compas.utilities import pairwise
from compas.geometry import centroid_points
from compas.geometry import is_coplanar
from compas.colors import Color
from .bufferobject import BufferObject


class MeshObject(BufferObject):
    """Object for displaying COMPAS mesh data structures.

    Parameters
    ----------
    data : :class: `compas.datastructures.Mesh`
        Mesh for the viewer
    name : string
        name of the object
    show_vertices : bool
        True to show vertices
    show_edges : bool
        True to show edges
    show_faces : bool
        True to show faces
    facecolor : list
        Face color
    linecolor : list
        Line color
    pointcolor : list
        point color
    linewidth : float
        Line width
    pointsize : float
        Point size
    hide_coplanaredges : bool
        True to hide the coplanar edges
    opacity : float
        The opacity of mesh
    vertices : list
        Subset of vertices to be displayed
    edges : list
        Subset of edges to be displayed
    faces : list
        Subset of faces to be displayed

    Attributes
    ----------
    facecolor : list
        Face color
    linecolor : list
        Line color
    pointcolor : list
        Face color
    linewidth : float
        Line width
    pointsize : float
        Point size
    hide_coplanaredges : bool
        True to hide the coplanar edges
    opacity : float
        The opacity of mesh
    vertices : list
        Subset of vertices to be displayed
    edges : list
        Subset of edges to be displayed
    faces : list
        Subset of faces to be displayed

    """

    def __init__(self, data, vertices=None, edges=None, faces=None,
                 hide_coplanaredges=False, use_vertex_color=False, **kwargs):
        super().__init__(data,  **kwargs)
        self._mesh = data
        self.hide_coplanaredges = hide_coplanaredges
        self.use_vertex_color = use_vertex_color
        self.vertices = vertices
        self.edges = edges
        self.faces = faces

    def _points_data(self):
        mesh = self._mesh
        vertex_xyz = {vertex: mesh.vertex_attributes(vertex, 'xyz') for vertex in mesh.vertices()}
        positions = []
        colors = []
        elements = []
        i = 0
        vertices = self.vertices or mesh.vertices()
        for vertex in vertices:
            positions.append(vertex_xyz[vertex])
            colors.append(self.pointcolors.get(vertex, self.pointcolor))
            elements.append([i])
            i += 1
        return positions, colors, elements

    def _lines_data(self):
        mesh = self._mesh
        vertex_xyz = {vertex: mesh.vertex_attributes(vertex, 'xyz') for vertex in mesh.vertices()}
        positions = []
        colors = []
        elements = []
        i = 0
        edges = self.edges or mesh.edges()
        for u, v in edges:
            color = self.linecolors.get((u, v), self.linecolor)
            if self.hide_coplanaredges:
                # hide the edge if neighbor faces are coplanar
                fkeys = mesh.edge_faces(u, v)
                if not mesh.is_edge_on_boundary(u, v):
                    ps = [mesh.face_center(fkeys[0]),
                          mesh.face_center(fkeys[1]),
                          *mesh.edge_coordinates(u, v)]
                    if is_coplanar(ps, tol=1e-5):
                        continue
            positions.append(vertex_xyz[u])
            positions.append(vertex_xyz[v])
            colors.append(color)
            colors.append(color)
            elements.append([i + 0, i + 1])
            i += 2
        return positions, colors, elements

    def _frontfaces_data(self):
        mesh = self._mesh
        vertex_xyz = {vertex: mesh.vertex_attributes(vertex, 'xyz') for vertex in mesh.vertices()}
        if self.use_vertex_color:
            vertex_color = {vertex: mesh.vertex_attribute(vertex, 'color') or Color.grey() for vertex in mesh.vertices()}
        positions = []
        colors = []
        elements = []
        i = 0
        faces = self.faces or mesh.faces()
        for face in faces:
            vertices = mesh.face_vertices(face)
            color = self.facecolors.get(face, self.facecolor)
            if len(vertices) == 3:
                a, b, c = vertices
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[b])
                positions.append(vertex_xyz[c])
                if self.use_vertex_color:
                    colors.append(vertex_color[a])
                    colors.append(vertex_color[b])
                    colors.append(vertex_color[c])
                else:
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                elements.append([i + 0, i + 1, i + 2])
                i += 3
            elif len(vertices) == 4:
                a, b, c, d = vertices
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[b])
                positions.append(vertex_xyz[c])
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[c])
                positions.append(vertex_xyz[d])
                if self.use_vertex_color:
                    colors.append(vertex_color[a])
                    colors.append(vertex_color[b])
                    colors.append(vertex_color[c])
                    colors.append(vertex_color[a])
                    colors.append(vertex_color[c])
                    colors.append(vertex_color[d])
                else:
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                elements.append([i + 0, i + 1, i + 2])
                elements.append([i + 3, i + 4, i + 5])
                i += 6
            else:
                points = [vertex_xyz[vertex] for vertex in vertices]
                c = centroid_points(points)
                for a, b in pairwise(points + points[:1]):
                    positions.append(a)
                    positions.append(b)
                    positions.append(c)
                    if self.use_vertex_color:
                        colors.append(vertex_color[a])
                        colors.append(vertex_color[b])
                        colors.append(vertex_color[c])
                    else:
                        colors.append(color)
                        colors.append(color)
                        colors.append(color)
                    elements.append([i + 0, i + 1, i + 2])
                    i += 3
        return positions, colors, elements

    def _backfaces_data(self):
        mesh = self._mesh
        vertex_xyz = {vertex: mesh.vertex_attributes(vertex, 'xyz') for vertex in mesh.vertices()}
        if self.use_vertex_color:
            vertex_color = {vertex: mesh.vertex_attribute(vertex, 'color') or Color.grey() for vertex in mesh.vertices()}
        positions = []
        colors = []
        elements = []
        i = 0
        faces = self.faces or mesh.faces()
        for face in faces:
            vertices = mesh.face_vertices(face)[::-1]
            color = self.facecolors.get(face, self.facecolor)
            if len(vertices) == 3:
                a, b, c = vertices
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[b])
                positions.append(vertex_xyz[c])
                if self.use_vertex_color:
                    colors.append(vertex_color[a])
                    colors.append(vertex_color[b])
                    colors.append(vertex_color[c])
                else:
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                elements.append([i + 0, i + 1, i + 2])
                i += 3
            elif len(vertices) == 4:
                a, b, c, d = vertices
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[b])
                positions.append(vertex_xyz[c])
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[c])
                positions.append(vertex_xyz[d])
                if self.use_vertex_color:
                    colors.append(vertex_color[a])
                    colors.append(vertex_color[b])
                    colors.append(vertex_color[c])
                    colors.append(vertex_color[a])
                    colors.append(vertex_color[c])
                    colors.append(vertex_color[d])
                else:
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                elements.append([i + 0, i + 1, i + 2])
                elements.append([i + 3, i + 4, i + 5])
                i += 6
            else:
                points = [vertex_xyz[vertex] for vertex in vertices]
                c = centroid_points(points)
                for a, b in pairwise(points + points[:1]):
                    positions.append(a)
                    positions.append(b)
                    positions.append(c)
                    if self.use_vertex_color:
                        colors.append(vertex_color[a])
                        colors.append(vertex_color[b])
                        colors.append(vertex_color[c])
                    else:
                        colors.append(color)
                        colors.append(color)
                        colors.append(color)
                    elements.append([i + 0, i + 1, i + 2])
                    i += 3
        return positions, colors, elements



import time
from OpenGL import GL

from qtpy import QtCore
from qtpy import QtWidgets

from compas_view2.scene import Camera
from compas_view2.scene import Mouse
from compas_view2.objects import GridObject


class View(QtWidgets.QOpenGLWidget):
    """Base OpenGL view widget.

    Parameters
    ----------
    app: :class:`compas_view2.app.App`
        The parent application of the view.
    background_color: tuple[float, float, float, float], optional
        The background or "clear" color of the view.
        Default is ``(1.0, 1.0, 1.0, 1.0)``.
    selection_color: tuple[float, float, float], optional
        The highlight color of selected objects.
        Default is ``(1.0, 1.0, 0.0)``.
    mode: 'shaded' | 'ghosted', optional
        The display mode.
        Default is ``'shaded'``.
    show_grid: bool, optional
        Flag for turning the grid on or off.
        Default is ``True``, which turns the grid on.
    """

    FRONT = 1
    RIGHT = 2
    TOP = 3
    PERSPECTIVE = 4

    def __init__(self,
                 app,
                 background_color=(1, 1, 1, 1),
                 selection_color=(1.0, 1.0, 0.0),
                 mode='shaded',
                 show_grid=True):
        super().__init__()
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self._opacity = 1.0
        self._current = View.PERSPECTIVE
        self.shader_model = None
        self.app = app
        self.color = background_color
        self.mode = mode
        self.selection_color = selection_color
        self.show_grid = show_grid
        self.camera = Camera(self)
        self.mouse = Mouse()
        self.grid = GridObject(1, 10, 10)
        self.objects = {}
        self.keys = {"shift": False, "control": False}
        self._frames = 0
        self._now = time.time()

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        self._mode = mode
        if mode == 'ghosted':
            self._opacity = 0.7
        else:
            self._opacity = 1.0
        if self.shader_model:
            self.shader_model.bind()
            self.shader_model.uniform1f("opacity", self._opacity)
            self.shader_model.release()
            self.update()

    @property
    def current(self):
        return self._current

    @current.setter
    def current(self, current):
        self._current = current
        if self.shader_model:
            self.shader_model.bind()
            self.shader_model.uniform4x4("projection", self.camera.projection(self.app.width, self.app.height))
            self.shader_model.release()
            self.update()

    @property
    def opacity(self):
        return self._opacity

    def clear(self):
        """Clear the view."""
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)

    def initializeGL(self):
        """Initialize the OpenGL canvas.

        This implements the virtual funtion of the OpenGL widget.
        See the PySide2 docs [1]_ for more info.
        It sets the clear color of the view,
        and enables culling, depth testing, blending, point smoothing, and line smoothing.

        To extend the behaviour of this function,
        you can implement :meth:`~compas_view2.views.View.init`.

        References
        ----------
        .. [1] https://doc.qt.io/qtforpython-5.12/PySide2/QtWidgets/QOpenGLWidget.html#PySide2.QtWidgets.PySide2.QtWidgets.QOpenGLWidget.initializeGL

        """
        GL.glClearColor(* self.color)
        GL.glPolygonOffset(1.0, 1.0)
        GL.glEnable(GL.GL_POLYGON_OFFSET_FILL)
        GL.glEnable(GL.GL_CULL_FACE)
        GL.glCullFace(GL.GL_BACK)
        GL.glEnable(GL.GL_DEPTH_TEST)
        GL.glDepthFunc(GL.GL_LESS)
        GL.glEnable(GL.GL_BLEND)
        GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
        GL.glEnable(GL.GL_POINT_SMOOTH)
        GL.glEnable(GL.GL_LINE_SMOOTH)
        GL.glEnable(GL.GL_FRAMEBUFFER_SRGB)
        self.init()

    def init(self):
        pass

    def resizeGL(self, w, h):
        """Resize the OpenGL canvas.

        This implements the virtual funtion of the OpenGL widget.
        See the PySide2 docs [1]_ for more info.

        To extend the behaviour of this function,
        you can implement :meth:`~compas_view2.views.View.resize`.

        Parameters
        ----------
        w: float
            The width of the canvas.
        h: float
            The height of the canvas.

        References
        ----------
        .. [1] https://doc.qt.io/qtforpython-5.12/PySide2/QtWidgets/QOpenGLWidget.html#PySide2.QtWidgets.PySide2.QtWidgets.QOpenGLWidget.resizeGL

        """
        GL.glViewport(0, 0, w, h)
        self.app.width = w
        self.app.height = h
        self.resize(w, h)

    def resize(self, w, h):
        pass

    def paintGL(self):
        """Paint the OpenGL canvas.

        This implements the virtual funtion of the OpenGL widget.
        See the PySide2 docs [1]_ for more info.

        To extend the behaviour of this function,
        you can implement :meth:`~compas_view2.views.View.paint`.

        Notes
        -----
        This method also paints the instance map used by the selector to identify selected objects.
        The instance map is immediately cleared again, after which the real scene objects are drawn.

        References
        ----------
        .. [1] https://doc.qt.io/qtforpython-5.12/PySide2/QtWidgets/QOpenGLWidget.html#PySide2.QtWidgets.PySide2.QtWidgets.QOpenGLWidget.paintGL

        """
        self.clear()
        self.paint()
        self._frames += 1
        if time.time() - self._now > 1:
            self._now = time.time()
            self.app.fps(self._frames)
            self._frames = 0

    def paint(self):
        pass

    def paint_instances(self):
        pass

    def paint_plane(self):
        pass

    def mouseMoveEvent(self, event):
        """Callback for the mouse move event.

        This method registers selections, if the left button is pressed,
        and modifies the view (pan/rotate), if the right button is pressed.
        """
        if not self.isActiveWindow() or not self.underMouse():
            return
        # record mouse position
        self.mouse.pos = event.pos()
        # compute displacement
        dx = self.mouse.dx()
        dy = self.mouse.dy()
        # do a box selection
        # if left button + SHIFT
        if event.buttons() & QtCore.Qt.LeftButton:
            if self.keys["shift"] or self.keys["control"]:
                self.app.selector.perform_box_selection(self.mouse.pos.x(), self.mouse.pos.y())
            # record mouse position
            self.mouse.last_pos = event.pos()
            self.update()
        # change the view
        # if right bottom
        elif event.buttons() & QtCore.Qt.RightButton:
            if self.keys["shift"]:
                self.camera.pan(dx, dy)
            else:
                self.camera.rotate(dx, dy)
            # record mouse position
            self.mouse.last_pos = event.pos()
            self.update()

    def mousePressEvent(self, event):
        if not self.isActiveWindow() or not self.underMouse():
            return
        # start selecting
        # if left button
        if event.buttons() & QtCore.Qt.LeftButton:
            self.mouse.buttons['left'] = True
            if self.keys["shift"] or self.keys["control"]:
                self.app.selector.reset_box_selection(event.pos().x(), event.pos().y())
        # do nothing
        # if right button
        elif event.buttons() & QtCore.Qt.RightButton:
            self.mouse.buttons['right'] = True
        # recod mouse position
        self.mouse.last_pos = event.pos()
        self.update()

    def mouseReleaseEvent(self, event):
        if not self.isActiveWindow() or not self.underMouse():
            return
        # finalize selecting
        # if left button
        if event.button() == QtCore.Qt.MouseButton.LeftButton:
            self.mouse.buttons['left'] = False
            # select location on grid
            if self.app.selector.wait_for_selection_on_plane:
                self.app.selector.finish_selection_on_plane(event.pos().x(), event.pos().y())
            # trigger object selection
            else:
                self.app.selector.enabled = True
        # do nothing
        # if right button
        elif event.button() == QtCore.Qt.MouseButton.RightButton:
            self.mouse.buttons['right'] = False
        self.update()

    def wheelEvent(self, event):
        if not self.isActiveWindow() or not self.underMouse():
            return
        degrees = event.delta() / 8
        steps = degrees / 15
        self.camera.zoom(steps)
        self.update()

    def keyPressEvent(self, event):
        key = event.key()
        if key == QtCore.Qt.Key_Return or key == QtCore.Qt.Key_Enter:
            self.app.selector.finish_selection()
        if key == QtCore.Qt.Key_Shift:
            self.app.selector.mode = "multi"
            self.keys["shift"] = True
        if key == QtCore.Qt.Key_Control:
            self.app.selector.mode = "deselect"
            self.keys["control"] = True

    def keyReleaseEvent(self, event):
        key = event.key()
        if key == QtCore.Qt.Key_Shift:
            self.app.selector.mode = self.app.selector.overwrite_mode or "single"
            self.keys["shift"] = False
        if key == QtCore.Qt.Key_Control:
            self.app.selector.mode = self.app.selector.overwrite_mode or "single"
            self.keys["control"] = False



from .flow import Flow  # noqa: F401
from .node import Node  # noqa: F401
from .values import ValueNode  # noqa: F401
from .values import IntegerNode  # noqa: F401
from .values import FloatNode  # noqa: F401
from .values import ChoiceNode  # noqa: F401
from .values import StringNode  # noqa: F401
from .values import VectorNode  # noqa: F401



from .button import Button  # noqa : F401
from .slider import Slider  # noqa : F401
from .radio import Radio  # noqa : F401
from .checkbox import Checkbox  # noqa : F401
from .select import Select  # noqa : F401



import numpy as np
from qtpy import QtWidgets

from compas.datastructures import Mesh
from compas.datastructures import Network
from compas.datastructures import VolMesh
from compas.colors import Color
from .collapsiblebox import CollapsibleBox


class PropertyForm(QtWidgets.QDockWidget):
    """Form class for real-time editing of objects

    Parameters
    ----------
    title: string
        The title of the form

    Attributes
    ----------
    on_update: function
        the function to be called when object attributes are updated from the form
    """

    def __init__(self, title, obj=None, on_update=None):
        super().__init__(title)

        if obj:
            self.set_object(obj, on_update)
        elif on_update:
            self.on_update = on_update

    def set_object(self, obj, on_update=None):
        scroll = QtWidgets.QScrollArea()
        self.setWidget(scroll)
        content = QtWidgets.QWidget()
        scroll.setWidget(content)
        scroll.setWidgetResizable(True)
        self._inputs = QtWidgets.QVBoxLayout(content)
        self.obj = obj
        if on_update:
            self.on_update = on_update

        # Show object class
        self.add_label(obj.name)
        self.add_label(obj._data.__class__)

        # Map object transform
        self.map_transform(obj)

        # Map object visualisation settings
        cb = self.add_collapsiblebox("Visualisation")
        v_layout = QtWidgets.QVBoxLayout()
        v_layout._parent = cb

        cb.setContentLayout(v_layout)
        self.map_object(obj, obj.visualisation, layout=v_layout)

        # Map custom object properties
        if obj.properties:
            cb = self.add_collapsiblebox("Object")
            v_layout = QtWidgets.QVBoxLayout()
            v_layout._parent = cb
            self.map_object(obj, obj.properties, layout=v_layout, update_data=True)
            cb.setContentLayout(v_layout)

        # Map object data
        if hasattr(obj._data, "data"):
            cb = self.add_collapsiblebox("Data")
            v_layout = QtWidgets.QVBoxLayout()
            v_layout._parent = cb
            self.data = obj._data.data
            if type(obj._data) in [Mesh, Network, VolMesh]:
                self.add_label(type(obj._data).__name__, layout=v_layout)
            else:
                self.map_inputs(self.data, layout=v_layout, update_data=True)
            cb.setContentLayout(v_layout)

        self._inputs.addStretch()

    def add_label(self, text, layout=None):
        if not layout:
            layout = QtWidgets.QHBoxLayout()
            self._inputs.addLayout(layout)
        # layout.setContentsMargins(0, 0, 0, 0)
        label = QtWidgets.QLabel(str(text))
        layout.addWidget(label)

    def add_collapsiblebox(self, name, layout=None):
        if hasattr(layout, "_parent"):
            cb = CollapsibleBox(name, parent=layout._parent)
        else:
            cb = CollapsibleBox(name)
        layout = layout or self._inputs
        layout.addWidget(cb)
        return cb

    def map_transform(self, obj):
        """Map the transformation of an object"""
        cb = self.add_collapsiblebox("Transform")
        v_layout = QtWidgets.QVBoxLayout()

        self.add_label("translation", layout=v_layout)
        layout = QtWidgets.QHBoxLayout()
        v_layout.addLayout(layout)
        self.map_number(obj.translation, 0, name="x", layout=layout)
        self.map_number(obj.translation, 1, name="y", layout=layout)
        self.map_number(obj.translation, 2, name="z", layout=layout)

        self.add_label("rotation", layout=v_layout)
        layout = QtWidgets.QHBoxLayout()
        v_layout.addLayout(layout)
        self.map_number(obj.rotation, 0, name="x", layout=layout)
        self.map_number(obj.rotation, 1, name="y", layout=layout)
        self.map_number(obj.rotation, 2, name="z", layout=layout)

        self.add_label("scale", layout=v_layout)
        layout = QtWidgets.QHBoxLayout()
        v_layout.addLayout(layout)
        self.map_number(obj.scale, 0, name="x", layout=layout)
        self.map_number(obj.scale, 1, name="y", layout=layout)
        self.map_number(obj.scale, 2, name="z", layout=layout)

        cb.setContentLayout(v_layout)

    def map_inputs(self, data, name=None, layout=None, update_data=False):
        """Map inputs of supported data type"""
        if name:
            self.add_label(name, layout=layout)
        if isinstance(data, (list, np.ndarray)):
            if isinstance(data[0], (float, int)):
                self.map_list(data, layout=layout, update_data=update_data)
            elif len(data) <= 10 and isinstance(data[0][0], (float, int)):
                self.map_vector_list(data, layout=layout, update_data=update_data)
            else:
                self.add_label(f"{data[0].__class__.__name__}[{len(data)}]", layout=layout)
        elif isinstance(data, Color):
            self.map_color(data, layout=layout, update_data=update_data)
        elif isinstance(data, dict):
            self.map_dict(data, layout=layout, update_data=update_data)
        else:
            raise TypeError("Un-supported data type")

    def map_object(self, data, attrs, layout=None, update_data=False):
        """Map the attributs of an object"""
        for attr in attrs:
            attribute = getattr(data, attr)
            if isinstance(attribute, bool):
                self.map_bool(data, attr, layout=layout, update_data=update_data)
            elif isinstance(attribute, (int, float)):
                self.map_number(data, attr, layout=layout, update_data=update_data)
            else:
                cb = self.add_collapsiblebox(attr, layout=layout)
                v_layout = QtWidgets.QVBoxLayout()
                v_layout._parent = cb
                self.map_inputs(attribute, layout=v_layout, update_data=update_data)
                cb.setContentLayout(v_layout)

    def map_dict(self, data, layout=None, update_data=False):
        """Map a dictionary input"""
        for key in data:
            if isinstance(data[key], (float, int)):
                self.map_number(data, key, layout=layout, update_data=update_data)
            else:
                cb = self.add_collapsiblebox(key, layout=layout)
                v_layout = QtWidgets.QVBoxLayout()
                v_layout._parent = cb
                self.map_inputs(data[key], layout=v_layout, update_data=update_data)
                cb.setContentLayout(v_layout)

    def map_vector_list(self, _list, layout=None, update_data=False):
        """Map a list of vetors"""
        v_layout = QtWidgets.QVBoxLayout()
        v_layout.setContentsMargins(20, 0, 0, 0)
        layout.addLayout(v_layout)
        for i, vector in enumerate(_list):
            self.map_list(vector, name=i, layout=v_layout, update_data=update_data)

    def map_list(self, _list, name=None, layout=None, update_data=False):
        """Map a list of numbers"""
        h_layout = QtWidgets.QHBoxLayout()
        layout = layout or self._inputs
        layout.addLayout(h_layout)
        if name is not None:
            self.add_label(str(name) + ": ", layout=h_layout)
        for i in range(len(_list)):
            self.map_number(_list, i, layout=h_layout, update_data=update_data)

    def map_number(self, obj, attribute, name=None, layout=None, update_data=False, minimum=float('-inf'), maximum=float('inf'), step=None):
        """Map number input field to an object attribute

        Parameters
        ----------
        obj: Object
            object to be edited
        attribute: string
            the name of attribute to be mapped

        Returns
        -------
        None
        """
        if not layout:
            layout = QtWidgets.QHBoxLayout()
            self._inputs.addLayout(layout)

        # layout.setContentsMargins(0, 0, 0, 0)
        label = QtWidgets.QLabel(name or str(attribute))
        if type(obj) in [list, dict, np.ndarray]:
            value = obj[attribute]
        else:
            value = getattr(obj, attribute)
        if isinstance(value, float):
            _input = QtWidgets.QDoubleSpinBox()
            _input.setSingleStep(0.1)
            _input.setMinimum(minimum)
            _input.setMaximum(maximum)
        elif isinstance(value, int):
            _input = QtWidgets.QSpinBox()
            if minimum == float('-inf'):
                minimum = -10**9
            if maximum == float('inf'):
                maximum = 10**9
            _input.setMinimum(minimum)
            _input.setMaximum(maximum)
        else:
            raise ValueError()
        if step:
            _input.setSingleStep(step)
        _input.setValue(value)
        layout.addWidget(label)
        layout.addWidget(_input)

        def set_number(value):
            if type(obj) in [list, dict, np.ndarray]:
                obj[attribute] = value
            else:
                setattr(obj, attribute, value)
            self.update(update_data)

        _input.valueChanged.connect(set_number)
        return layout

    def map_color(self, color, layout=None, update_data=False):
        """Map color input field to an object attribute
        """
        h_layout = QtWidgets.QHBoxLayout()
        layout = layout or self._inputs
        layout.addLayout(h_layout)
        for channel in ["r", "g", "b"]:
            self.map_number(color, channel, layout=h_layout, update_data=update_data, minimum=0, maximum=1, step=0.01)

    def map_bool(self, obj, attribute, name=None, layout=None, update_data=False):
        """Map color input field to an object attribute
        """
        if not layout:
            layout = QtWidgets.QHBoxLayout()
            self._inputs.addLayout(layout)

        if type(obj) in [list, dict, np.ndarray]:
            value = obj[attribute]
        else:
            value = getattr(obj, attribute)
        _input = QtWidgets.QCheckBox(name or str(attribute))
        _input.setChecked(value)
        layout.addWidget(_input)

        def set_bool(value):
            if type(obj) in [list, dict, np.ndarray]:
                obj[attribute] = value
            else:
                setattr(obj, attribute, value)
            self.update(update_data)

        _input.stateChanged.connect(set_bool)
        return layout

    def update(self, update_data):
        if update_data:
            if hasattr(self, "data"):
                try:
                    self.obj._data.data = self.data
                except Exception as e:
                    print(e)
                    print("Failed to update data of", self.obj)
            self.obj.update()
        else:
            self.obj._update_matrix()

        if self.on_update:
            self.on_update()

    def inputs(self):
        """Creates layout for input fields

        Returns
        -------
        QtWidgets.QVBoxLayout
        """
        return QtWidgets.QVBoxLayout()



from compas.utilities import pairwise
from compas.geometry import centroid_points
from .bufferobject import BufferObject
from compas_occ.brep import BRep


class BRepObject(BufferObject):
    """Object for displaying compas_occ Brep.

    Attributes
    ----------
    brep : :class:`compas_occ.brep.BRep`
        The compas_occ Brep object.
    mesh : :class:`compas.datastructures.Mesh`
        The tesselation mesh representation of the Brep.
    """

    def __init__(self, brep: BRep, **kwargs):
        super().__init__(brep, **kwargs)
        mesh, boundaries = brep.to_viewmesh()
        self._mesh = mesh
        self._boundaries = boundaries

    @property
    def brep(self):
        return self.item

    @property
    def mesh(self):
        return self._mesh

    @property
    def boundaries(self):
        return self._boundaries

    def update(self):
        mesh, boundaries = self.brep.to_viewmesh()
        self._mesh = mesh
        self._boundaries = boundaries
        super().update()

    def _lines_data(self):
        positions = []
        colors = []
        elements = []
        lines = []
        for polyline in self.boundaries:
            lines += pairwise(polyline.points)
        count = 0
        for i, (pt1, pt2) in enumerate(lines):
            positions.append(pt1)
            positions.append(pt2)
            color = self.linecolors.get(i, self.linecolor)
            colors.append(color)
            colors.append(color)
            elements.append([count, count + 1])
            count += 2
        return positions, colors, elements

    def _frontfaces_data(self):
        mesh = self.mesh
        vertex_xyz = {
            vertex: mesh.vertex_attributes(vertex, "xyz") for vertex in mesh.vertices()
        }
        positions = []
        colors = []
        elements = []
        i = 0
        faces = mesh.faces()
        for face in faces:
            vertices = mesh.face_vertices(face)
            color = self.facecolors.get(face, self.facecolor)
            if len(vertices) == 3:
                a, b, c = vertices
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[b])
                positions.append(vertex_xyz[c])
                colors.append(color)
                colors.append(color)
                colors.append(color)
                elements.append([i + 0, i + 1, i + 2])
                i += 3
            elif len(vertices) == 4:
                a, b, c, d = vertices
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[b])
                positions.append(vertex_xyz[c])
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[c])
                positions.append(vertex_xyz[d])
                colors.append(color)
                colors.append(color)
                colors.append(color)
                colors.append(color)
                colors.append(color)
                colors.append(color)
                elements.append([i + 0, i + 1, i + 2])
                elements.append([i + 3, i + 4, i + 5])
                i += 6
            else:
                points = [vertex_xyz[vertex] for vertex in vertices]
                c = centroid_points(points)
                for a, b in pairwise(points + points[:1]):
                    positions.append(a)
                    positions.append(b)
                    positions.append(c)
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                    elements.append([i + 0, i + 1, i + 2])
                    i += 3

        return positions, colors, elements

    def _backfaces_data(self):
        mesh = self._mesh
        vertex_xyz = {
            vertex: mesh.vertex_attributes(vertex, "xyz") for vertex in mesh.vertices()
        }
        positions = []
        colors = []
        elements = []
        i = 0
        faces = mesh.faces()
        for face in faces:
            vertices = mesh.face_vertices(face)[::-1]
            color = self.facecolors.get(face, self.facecolor)
            if len(vertices) == 3:
                a, b, c = vertices
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[b])
                positions.append(vertex_xyz[c])
                colors.append(color)
                colors.append(color)
                colors.append(color)
                elements.append([i + 0, i + 1, i + 2])
                i += 3
            elif len(vertices) == 4:
                a, b, c, d = vertices
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[b])
                positions.append(vertex_xyz[c])
                positions.append(vertex_xyz[a])
                positions.append(vertex_xyz[c])
                positions.append(vertex_xyz[d])
                colors.append(color)
                colors.append(color)
                colors.append(color)
                colors.append(color)
                colors.append(color)
                colors.append(color)
                elements.append([i + 0, i + 1, i + 2])
                elements.append([i + 3, i + 4, i + 5])
                i += 6
            else:
                points = [vertex_xyz[vertex] for vertex in vertices]
                c = centroid_points(points)
                for a, b in pairwise(points + points[:1]):
                    positions.append(a)
                    positions.append(b)
                    positions.append(c)
                    colors.append(color)
                    colors.append(color)
                    colors.append(color)
                    elements.append([i + 0, i + 1, i + 2])
                    i += 3
        return positions, colors, elements



from qtpy import QtCore
from qtpy import QtWidgets


class CollapsibleBox(QtWidgets.QWidget):
    def __init__(self, title="", parent=None):
        super(CollapsibleBox, self).__init__(parent)

        self._parent = parent
        self._expanded = False
        self.toggle_button = QtWidgets.QToolButton(
            text=title, checkable=True, checked=False
        )
        self.toggle_button.setStyleSheet("QToolButton { border: none; }")
        self.toggle_button.setToolButtonStyle(
            QtCore.Qt.ToolButtonTextBesideIcon
        )
        self.toggle_button.setIconSize(QtCore.QSize(8, 8))
        self.toggle_button.setArrowType(QtCore.Qt.RightArrow)
        self.toggle_button.pressed.connect(self.on_pressed)

        self.content_area = QtWidgets.QScrollArea(
            maximumHeight=0, minimumHeight=0
        )
        self.content_area.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed
        )
        # self.content_area.setFrameShape(QtWidgets.QFrame.NoFrame)
        lay = QtWidgets.QVBoxLayout(self)
        lay.setSpacing(0)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.addWidget(self.toggle_button)
        lay.addWidget(self.content_area)

    def update_parent(self):
        if self._parent:
            self._parent.expand()
            self._parent.update_parent()

    def on_pressed(self):
        self._expanded = not self.toggle_button.isChecked()
        if self._expanded:
            self.expand()
        else:
            self.collapse()
        self.update_parent()

    def get_content_height(self):
        return self._layout.sizeHint().height()

    def setContentLayout(self, layout):
        self._layout = layout
        self.content_area.setLayout(layout)
        self._collapsed_height = self.sizeHint().height()

    def expand(self):
        self.toggle_button.setArrowType(QtCore.Qt.DownArrow)
        self.setFixedHeight(self._collapsed_height + self.get_content_height())
        self.content_area.setMaximumHeight(self.get_content_height())

    def collapse(self):
        self.toggle_button.setArrowType(QtCore.Qt.RightArrow)
        self.setFixedHeight(self._collapsed_height)
        self.content_area.setMaximumHeight(0)


if __name__ == "__main__":
    pass



from qtpy import QtCore


class Mouse:
    """Class representing mouse actions and movements.

    Attributes
    ----------
    pos : QtCore.QPoint
        The current position of the mouse on the screen.
    last_pos : QtCore.QPoint
        The last recorded position of the mouse on the screen.
    buttons : dict
        A dict containing keys "left" and "right"
        with each an associated boolean indicating if the corresponding mouse button is pressed.

    """

    def __init__(self):
        self.pos = QtCore.QPoint()
        self.last_pos = QtCore.QPoint()
        self.buttons = {'left': False, 'right': False}

    def dx(self):
        return self.pos.x() - self.last_pos.x()

    def dy(self):
        return self.pos.y() - self.last_pos.y()



from .bufferobject import BufferObject


class NetworkObject(BufferObject):
    """Object for displaying COMPAS network data structures."""

    def __init__(self, data, show_points: bool = True, **kwargs):
        super().__init__(data, show_points=show_points, **kwargs)

    @property
    def nodes(self):
        return self._nodes

    @property
    def edges(self):
        return self._edges

    def _points_data(self):
        data = self._data
        node_xyz = {node: data.node_attributes(node, 'xyz') for node in data.nodes()}
        # nodes
        positions = []
        colors = []
        elements = []
        for i, node in enumerate(data.nodes()):
            positions.append(node_xyz[node])
            colors.append(self.pointcolors.get(node, self.pointcolor))
            elements.append([i])
        return positions, colors, elements

    def _lines_data(self):
        data = self._data
        node_xyz = {node: data.node_attributes(node, 'xyz') for node in data.nodes()}
        positions = []
        colors = []
        elements = []
        i = 0
        for u, v in data.edges():
            positions.append(node_xyz[u])
            positions.append(node_xyz[v])
            color = self.linecolors.get((u, v), self.linecolor)
            colors.append(color)
            colors.append(color)
            elements.append([i + 0, i + 1])
            i += 2
        return positions, colors, elements



from .dockform import DockForm
from .treeform import TreeForm
from qtpy import QtWidgets


class TabsForm(DockForm):
    def __init__(self, app, title="Tree", tabs=[], columns=["key", "value"], show_headers=True, striped_rows=False):
        super().__init__(app, title)
        self.treeforms = {}
        containerwidget = QtWidgets.QWidget()
        tabwidget = QtWidgets.QTabWidget()
        for tab in tabs:
            treeForm = TreeForm(app, title=tab["name"], data=tab["data"], columns=tab.get("columns") or columns, show_headers=show_headers, striped_rows=striped_rows)
            treeForm.setTitleBarWidget(QtWidgets.QWidget())
            tabwidget.addTab(treeForm, tab["name"])
            self.treeforms[tab["name"]] = treeForm
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(tabwidget)
        containerwidget.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setWidget(containerwidget)

    def update(self, tabs):
        for tab in tabs:
            self.treeforms[tab["name"]].update(tab["data"])



from compas.geometry import Vector
from compas.geometry import Shape


class VectorGroup(Shape):

    def __init__(self, vectors, positions=None):
        super().__init__()
        print(vectors[0])
        self.vectors = [Vector(*v) for v in vectors]
        if positions:
            self.positions = [Vector(*p) for p in positions]
        else:
            self.positions = [Vector(0, 0, 0) for _ in range(len(vectors))]

    @property
    def data(self):
        return {'positions': [list(p) for p in self.positions], 'vectors': [list(d) for d in self.vectors], 'sizes': self.sizes, 'colors': self.colors}

    @data.setter
    def data(self, data):
        self.positions = [Vector(*p) for p in data['positions']]
        self.vectors = [Vector(*d) for d in data['vectors']]
        self.colors = data['colors']
        self.sizes = data['sizes']

    # ==========================================================================
    # customisation
    # ==========================================================================

    def __repr__(self):
        return 'VectorGroup[{}]'.format(len(self.positions))

    # ==========================================================================
    # constructors
    # ==========================================================================

    @classmethod
    def from_data(cls, data):
        vectorgroup = cls()
        vectorgroup.data = data
        return vectorgroup

    # ==========================================================================
    # methods
    # ==========================================================================
    def transform(self, transformation):
        for p in self.positions:
            p.transform(transformation)
        for d in self.vectors:
            d.transform(transformation)



from qtpy import QtWidgets


class DockForm(QtWidgets.QDockWidget):
    """
    """

    def __init__(self, app, title):
        super().__init__(title)
        self.app = app
        self.setMinimumWidth(200)
        scroll = QtWidgets.QScrollArea()
        self.setWidget(scroll)
        content = QtWidgets.QWidget()
        scroll.setWidget(content)
        scroll.setWidgetResizable(True)
        vlay = QtWidgets.QVBoxLayout(content)
        self.content_layout = vlay

    def clear(self):
        for i in reversed(range(self.content_layout.count())):
            widget = self.content_layout.itemAt(i).widget()
            if widget is not None:
                widget.deleteLater()



from math import tan
from math import radians

from compas.geometry import Transformation
from compas.geometry import normalize_vector
from compas.geometry import subtract_vectors
from compas.geometry import cross_vectors


def ortho(left, right, bottom, top, near, far):
    """Construct an orthogonal projection matrix.

    Parameters
    ----------
    left : float
        Location of the left clipping plane.
    right : float
        Location of the right clipping plane.
    bottom : float
        Location of the bottom clipping plane.
    top : float
        Location of the top clipping plane.
    near : float
        Location of the near clipping plane.
    far : float
        Location of the far clipping plane.

    Returns
    -------
    :class:`compas.geometry.Transformation`

    """
    dx = right - left
    dy = top - bottom
    dz = far - near
    rx = -(right + left) / dx
    ry = -(top + bottom) / dy
    rz = -(far + near) / dz
    matrix = [
        [2.0 / dx,        0,         0, rx],
        [       0, 2.0 / dy,         0, ry],  # noqa: E201
        [       0,        0, -2.0 / dz, rz],  # noqa: E201
        [       0,        0,         0,  1]   # noqa: E201
    ]
    return Transformation.from_matrix(matrix)


def perspective(fov, aspect, near, far):
    """Construct a perspective projection matrix.

    Parameters
    ----------
    fov : float
        The field of view in degrees.
    aspect : float
        The aspect ratio of the view.
    near : float
        Location of the near clipping plane.
    far : float
        Location of the far clipping plane.

    Returns
    -------
    :class:`compas.geometry.Transformation`

    """
    sy = 1.0 / tan(radians(fov) / 2.0)
    sx = sy / aspect
    zz = (far + near) / (near - far)
    zw = 2 * far * near / (near - far)
    matrix = [
        [sx,  0,  0,  0],
        [ 0, sy,  0,  0],  # noqa: E201
        [ 0,  0, zz, zw],  # noqa: E201
        [ 0,  0, -1,  0]   # noqa: E201
    ]
    return Transformation.from_matrix(matrix)


def lookat(eye, target, up):
    """Construct a "look at" transformation matrix.

    Parameters
    ----------
    eye : [float, float, float] or :class:`compas.geometry.Point`
        Location of the "eye" of the camera.
    target : [float, float, float] or :class:`compas.geometry.Point`
        Location of the target to look at.
    up : [float, float, float] or :class:`compas.geometry.Vector`
        Direction indicating which way is "up".

    Returns
    -------
    :class:`compas.geometry.Transformation`

    """
    d = normalize_vector(subtract_vectors(target, eye))
    r = cross_vectors(d, normalize_vector(up))
    u = cross_vectors(r, d)
    matrix = [
        [+r[0], +r[1], +r[2], -eye[0]],
        [+u[0], +u[1], +u[2], -eye[1]],
        [-d[0], -d[1], -d[2], -eye[2]],
        [    0,     0,    0,       1]  # noqa: E201
    ]
    return Transformation.from_matrix(matrix)



import os
import numpy as np
import freetype as ft
from OpenGL import GL

from compas_view2.gl import make_index_buffer
from compas_view2.gl import make_vertex_buffer

from .object import Object

here = os.path.dirname(__file__)
fonts = os.path.join(here, '..', 'fonts')


class TextObject(Object):
    """Object for displaying text sprites."""

    def __init__(self, data, color=None, height=10, **kwargs):
        super().__init__(data, **kwargs)
        self.color = color or [0, 0, 0]
        self.characters = []
        self.buffers = []
        self.height = height

    def init(self):
        self.make_buffers()

    def make_buffers(self):
        self._text_buffer = {
            'positions': make_vertex_buffer(self._data.position),
            'elements': make_index_buffer([0]),
            'text_texture': self.make_text_texture(),
            'n': 1
        }

    def make_text_texture(self):
        # change the filename if necessary
        face = ft.Face(os.path.join(fonts, 'FreeSans.ttf'))

        char_width = 48
        char_height = 80
        # the size is specified in 1/64 pixel
        face.set_char_size(64*char_width)

        text = self._data.text
        string_buffer = np.zeros(shape=(char_height, char_width*len(text)))

        for i, c in enumerate(text):
            if c == " ":
                continue
            face.load_char(c, ft.FT_LOAD_FLAGS['FT_LOAD_RENDER'])
            glyph = face.glyph
            bitmap = glyph.bitmap
            char = np.array(bitmap.buffer)
            char = char.reshape((bitmap.rows, bitmap.width))
            string_buffer[-char.shape[0]:, i*char_width: i*char_width+char.shape[1]] = char

        string_buffer = string_buffer.reshape((string_buffer.shape[0]*string_buffer.shape[1]))

        # create glyph texture
        texture = GL.glGenTextures(1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, texture)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)
        GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_R8, char_width*len(text), char_height, 0, GL.GL_RED, GL.GL_UNSIGNED_BYTE, string_buffer)
        return texture

    def draw(self, shader):
        """Draw the object from its buffers"""
        shader.enable_attribute('position')
        shader.uniform4x4('transform', self.matrix)
        shader.uniform1f('object_opacity', self.opacity)
        shader.uniform1i('text_height', self._data.height)
        shader.uniform1i('text_num', len(self._data.text))
        shader.uniform3f('text_color', self.color)
        shader.uniformTex("text_texture", self._text_buffer['text_texture'])
        shader.bind_attribute('position', self._text_buffer['positions'])
        shader.draw_texts(elements=self._text_buffer['elements'], n=self._text_buffer['n'])
        shader.uniform1i('is_text', 0)
        shader.uniform1f('object_opacity', 1)
        shader.disable_attribute('position')



from qtpy import QtWidgets
from qtpy import QtCore

from .form import Form


class LineForm(Form):

    def __init__(self):
        super().__init__(title='Add Line')

    def float_field(self, text, value, layout=None, parent=None):
        if not layout:
            layout = QtWidgets.QHBoxLayout()
        label = QtWidgets.QLabel(text)
        field = QtWidgets.QDoubleSpinBox()
        field.setValue(value)
        layout.addWidget(label)
        layout.addWidget(field)
        if parent:
            parent.addLayout(layout)
        return field

    def inputs(self):
        inputs = QtWidgets.QVBoxLayout()
        points = QtWidgets.QHBoxLayout()
        # point A
        A_group = QtWidgets.QGroupBox('Point A')
        A_layout = QtWidgets.QVBoxLayout()
        A_group.setLayout(A_layout)
        self.Ax_input = self.float_field('X', 0.0, parent=A_layout)
        self.Ay_input = self.float_field('Y', 0.0, parent=A_layout)
        self.Az_input = self.float_field('Z', 0.0, parent=A_layout)
        # point B
        B_group = QtWidgets.QGroupBox('Point B')
        B_layout = QtWidgets.QVBoxLayout()
        B_group.setLayout(B_layout)
        self.Bx_input = self.float_field('X', 0.0, parent=B_layout)
        self.By_input = self.float_field('Y', 0.0, parent=B_layout)
        self.Bz_input = self.float_field('Z', 0.0, parent=B_layout)
        # show points
        self.show_points_input = QtWidgets.QCheckBox('Show Points')
        self.show_points_input.setCheckState(QtCore.Qt.Checked)
        # combine
        points.addWidget(A_group)
        points.addWidget(B_group)
        inputs.addLayout(points)
        inputs.addWidget(self.show_points_input)
        return inputs

    @property
    def Ax(self):
        return float(self.Ax_input.value())

    @property
    def Ay(self):
        return float(self.Ay_input.value())

    @property
    def Az(self):
        return float(self.Az_input.value())

    @property
    def Bx(self):
        return float(self.Bx_input.value())

    @property
    def By(self):
        return float(self.By_input.value())

    @property
    def Bz(self):
        return float(self.Bz_input.value())

    @property
    def show_points(self):
        return bool(self.show_points_input.checkState())



from compas.datastructures import Mesh
from compas.geometry import Plane
from compas.geometry import Torus
from .meshobject import MeshObject


class TorusObject(MeshObject):
    """Object for displaying COMPAS torus geometry."""

    def __init__(self, data, u=16, v=16, **kwargs):
        super().__init__(Mesh.from_shape(data, u=u, v=v), **kwargs)
        self.u = u
        self.v = v
        self._data = data

    def update(self):
        self._mesh = Mesh.from_shape(self._data, u=self.u, v=self.v)
        self.init()
        super().update()

    @property
    def properties(self):
        return ["u", "v"]

    @classmethod
    def create_default(cls) -> Torus:
        return Torus(Plane([0, 0, 0], [0, 0, 1]), 1, 0.5)



from .editform import EditForm


class PointEditForm(EditForm):
    """Form class for real-time editing of PointObjects

    Parameters
    ----------
    pointobject: :class:`compas_view2.objects.PointObject`
        The point object to be edited
    on_update: function
        the function to be called when object attributes are updated from the form
    """

    def __init__(self, pointobject, on_update=None):
        super().__init__("Edit Point", on_update=on_update)
        self.map_number(pointobject._data, 'x')
        self.map_number(pointobject._data, 'y')
        self.map_number(pointobject._data, 'z')



import ctypes as ct
from OpenGL import GL


def gl_info():
    """Return formatted information about the current GL implementation.

    Returns
    -------
    str

    """
    info = """
        Vendor: {0}
        Renderer: {1}
        OpenGL Version: {2}
        Shader Version: {3}
        """.format(
        GL.glGetString(GL.GL_VENDOR),
        GL.glGetString(GL.GL_RENDERER),
        GL.glGetString(GL.GL_VERSION),
        GL.glGetString(GL.GL_SHADING_LANGUAGE_VERSION)
    )
    return info


def make_vertex_buffer(data, dynamic=False):
    """Make a vertex buffer from the given data.

    Parameters
    ----------
    data: list[float]
        A flat list of floats.
    dynamic : bool, optional
        If True, the buffer is optimized for dynamic access.

    Returns
    -------
    int
        Vertex buffer ID.

    Examples
    --------
    >>> from compas.utilities import flatten
    >>> vertices = [[0, 0, 0], [1, 0, 0], [2, 0, 0], [3, 0, 0]]
    >>> buffer = make_vertex_buffer(list(flatten(vertices)))

    """
    access = GL.GL_DYNAMIC_DRAW if dynamic else GL.GL_STATIC_DRAW
    n = len(data)
    size = n * ct.sizeof(ct.c_float)
    vbo = GL.glGenBuffers(1)
    data = (ct.c_float * n)(* data)
    GL.glBindBuffer(GL.GL_ARRAY_BUFFER, vbo)
    GL.glBufferData(GL.GL_ARRAY_BUFFER, size, data, access)
    GL.glBindBuffer(GL.GL_ARRAY_BUFFER, 0)
    return vbo


def make_index_buffer(data, dynamic=False):
    """Make an element buffer from the given data.

    Parameters
    ----------
    data: list[int]
        A flat list of ints.
    dynamic : bool, optional
        If True, the buffer is optimized for dynamic access.

    Returns
    -------
    int
        Element buffer ID.

    Examples
    --------
    >>> from compas.utilities import flatten
    >>> edges = [(0, 1), (1, 2), (2, 3)]
    >>> buffer = make_index_buffer(list(flatten(edges)))

    """
    access = GL.GL_DYNAMIC_DRAW if dynamic else GL.GL_STATIC_DRAW
    n = len(data)
    size = n * ct.sizeof(ct.c_uint)
    vbo = GL.glGenBuffers(1)
    data = (ct.c_int * n)(* data)
    GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, vbo)
    GL.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, size, data, access)
    GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0)
    return vbo


def update_vertex_buffer(data, buffer):
    """Update a vertex buffer with new data.

    Parameters
    ----------
    data: list[float]
        A flat list of floats.
    buffer : int
        The ID of the buffer.

    Returns
    -------
    None

    """
    n = len(data)
    size = n * ct.sizeof(ct.c_float)
    data = (ct.c_float * n)(* data)
    GL.glBindBuffer(GL.GL_ARRAY_BUFFER, buffer)
    GL.glBufferSubData(GL.GL_ARRAY_BUFFER, 0, size, data)
    GL.glBindBuffer(GL.GL_ARRAY_BUFFER, 0)


def update_index_buffer(data, buffer):
    """Update an index buffer with new data.

    Parameters
    ----------
    data: list[int]
        A flat list of ints.
    buffer : int
        The ID of the buffer.

    Returns
    -------
    None

    """
    n = len(data)
    size = n * ct.sizeof(ct.c_uint)
    data = (ct.c_int * n)(* data)
    GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, buffer)
    GL.glBufferSubData(GL.GL_ELEMENT_ARRAY_BUFFER, 0, size, data)
    GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0)



from qtpy.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QPushButton, QCheckBox, QLabel


class ExecutionControl(QWidget):
    """Main widget to control and monitor the execution of node."""

    def __init__(self, params):
        self.node, self.node_item = params
        QWidget.__init__(self)

        layout = QVBoxLayout()
        self.setStyleSheet("background: transparent; color: white;")
        self.setLayout(layout)

        h_layout = QHBoxLayout()
        layout.addLayout(h_layout)

        self.checkbox = QCheckBox('Auto Update')
        self.checkbox.setStyleSheet("color: white;")
        h_layout.addWidget(self.checkbox)

        self.button = QPushButton('Run')
        self.button.setStyleSheet("background-color: #0092D2;")
        h_layout.addWidget(self.button)

        self.message = QLabel()
        self.message.setMaximumWidth(200)
        layout.addWidget(self.message)
        self.set_message()

        self.pause_event = False

        if self.node.block_updates:
            self.button.setVisible(True)
            self.checkbox.setChecked(False)
        else:
            self.button.setVisible(False)
            self.checkbox.setChecked(True)

        self.button.clicked.connect(self.update_node)
        self.checkbox.stateChanged.connect(self.set_auto_update)

    def set_message(self, message=None):
        if message:
            self.message.setText(message)
            self.message.setVisible(True)
        else:
            self.message.setVisible(False)

    def set_auto_update(self, _, value=None, update_node=True):

        if self.pause_event:
            return

        if value is not None:
            self.pause_event = True
            self.checkbox.setChecked(value)
            self.pause_event = False

        auto_update = self.checkbox.isChecked()
        self.node.block_updates = not auto_update
        self.button.setVisible(not auto_update)
        self.node_item.update_shape()

        if update_node and auto_update:
            self.update_node()

    def update_node(self):
        self.node.update_event()

    def get_state(self) -> dict:
        data = {}
        return data

    def set_state(self, data: dict):
        pass

    def remove_event(self):
        pass



from qtpy import QtWidgets
from qtpy import QtCore


class Checkbox:
    """Add a checkbox.

    Parameters
    ----------
    app : :class:`compas_view2.app.App`
        The app containing the widget.
    parent : QtWidgets.QWidget
        The parent widget for the checkbox.
    text : str
        The text label of the checkbox.
    action : callable
        The action associated with the checkox.
    checked : bool, optional
        If True, the checkbox will be displayed as checked.

    Attributes
    ----------
    checkbox : QtWidgets.QCheckBox
        The actual checkbox widget.
    action : callable
        The action associated with the toggle event of the checkbox.

    """

    def __init__(self,
                 app,
                 parent,
                 *,
                 text,
                 action,
                 checked=False):

        box = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        checkbox = QtWidgets.QCheckBox(text)
        checkbox.setCheckState(QtCore.Qt.CheckState.Checked if checked else QtCore.Qt.CheckState.Unchecked)
        layout.addWidget(checkbox)
        box.setLayout(layout)
        parent.addWidget(box)
        # connect actions
        checkbox.toggled.connect(self)
        checkbox.toggled.connect(app.view.update)
        # identify attributes
        self.checkbox = checkbox
        self.action = action

    def __call__(self, *args, **kwargs):
        """Wrapper for the action associated with the checkbox.

        Returns
        -------
        None

        """
        self.action(args[0])
